"""Back-end for BigQuery-related code."""
import re
import copy
import typing as T
from datetime import datetime, timedelta
import json
from pathlib import Path
import warnings

from google.cloud import bigquery
from google.cloud.bigquery.job.query import QueryJob
from google.oauth2.service_account import Credentials
try:
    import pandas as pd
except ImportError:
    pass

from .carol import Carol
from .storage import Storage
from .connectors import Connectors
from . import __TEMP_STORAGE__

CACHE_FILE_NAME = '.pycarol_temp_{env_id}.json'


class BQ:

    token = None

    def __init__(self, carol: Carol, service_account: dict = None, cache_cds: bool = True):

        self.carol = carol
        self.service_account = service_account
        self._provided_sa = service_account is not None
        self.env = carol.get_current()
        self._temp_file_name = CACHE_FILE_NAME.format(
            env_id=self.env['env_id'])
        self._temp_file_path = Path(__TEMP_STORAGE__) / self._temp_file_name
        self.client = None
        self.dataset_id = f"carol-{self.env['env_id'][0:20]}.{self.env['env_id']}"
        self.cache_cds = cache_cds
        self.storage = None
        self._fetch_cache()

    def _format_sa(self, sa):

        expiration_estimate = datetime.strptime(
            sa['expiration_time'], '%Y-%m-%dT%H:%M:%S.%fZ')
        sa = {
            'sa': sa,
            'expiration_time': expiration_estimate,
            'env': copy.deepcopy(self.env)
        }
        return sa

    def _fetch_cache(self):

        if not self.is_expired():
            return
        if self._temp_file_path.exists():
            sa = self._load_local_cache()
            sa = self._format_sa(sa)
            if sa['expiration_time'] < datetime.utcnow():
                return None
            else:
                BQ.token = sa
                self.service_account = sa['sa']
                return

        if self.cache_cds:
            if not self.storage:
                self.storage = Storage(self.carol)
            if self.storage.exists(name=self._temp_file_name, storage_space='pycarol'):
                sa_file = self.storage.load(name=str(self._temp_file_name),
                                            format='file', storage_space='pycarol', cache=False)
                sa = self._load_local_cache(sa_file)
                sa = self._format_sa(sa)

                if sa['expiration_time'] < datetime.utcnow():
                    return None
                else:
                    BQ.token = sa
                    self.service_account = sa['sa']
                    self._save_local_cache()

    def _load_local_cache(self, local_cache: str = None):
        local_cache = local_cache or self._temp_file_path
        with open(local_cache, 'r') as f:
            sa = json.load(f)
        return sa

    def _save_local_cache(self):
        self._temp_file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._temp_file_path, "w") as f:
            json.dump(self.service_account, f)

    def _save_cache(self):

        self._save_local_cache()
        if self.cache_cds:
            self._save_cds_cache()

    def _save_cds_cache(self):
        if self._temp_file_path.exists():
            if self.storage is None:
                self.storage = Storage(self.carol)
            self.storage.save(name=self._temp_file_name, obj=str(self._temp_file_path),
                              format='file', storage_space='pycarol')

    def _generate_client(self) -> bigquery.Client:
        """Generate client from credentials."""

        if self._provided_sa:
            service_account = self.service_account
        else:
            service_account = BQ.token['sa']

        credentials = Credentials.from_service_account_info(service_account)
        project = service_account["project_id"]
        client = bigquery.Client(project=project, credentials=credentials)
        return client

    def is_expired(self):

        if self._provided_sa:
            return False
        elif BQ.token is None and self.service_account is None:
            return True
        elif (BQ.token is not None) and ((BQ.token['expiration_time'] < datetime.utcnow()) or (BQ.token['env']['env_id'] != self.env['env_id'])):
            return True
        elif BQ.token is None:
            return True
        else:
            return False

    def get_credential(self, expiration_time: int = 24, force: bool = False) -> dict:
        """ Get service account for BigQuery.

        Args:
            expiration_time (int): Time in hours for credentials to expire. Max value 24.
            force (bool): Force to get new credentials skiping any cache.

        Returns:
            dict: Service account

        .. code:: python


            from pycarol import Carol
            from pycarol.bigquery import BQ

            bq = BQ(Carol())
            service_account = bq.get_credential(expiration_time=120)


        """

        if force or self.is_expired():

            current_time = datetime.utcnow()
            url = 'v1/create_temporary_key'
            prefix_path = '/sql/v1/api/'
            env = self.carol.get_current()
            payload = {
                "expirationTime": expiration_time,
                "mdmOrgId": env['org_id'],
                "mdmTenantId":  env['env_id']
            }
            self.service_account = self.carol.call_api(
                method='POST', path=url, prefix_path=prefix_path, data=payload)

            if not 'expiration_time' in self.service_account:
                self.service_account['expiration_time'] = datetime.strftime(
                    current_time + timedelta(hours=expiration_time), '%Y-%m-%dT%H:%M:%S.%fZ',)

            BQ.token = self._format_sa(self.service_account)

            try:
                self._save_cache()
            except Exception as e:
                warnings.warn(
                    f"Failed to save cache {e}",
                    UserWarning, stacklevel=3
                )

        if self.service_account is None:
            self.service_account = BQ.token['sa']
        return self.service_account

    def query(
        self,
        query: str,
        dataset_id: T.Optional[str] = None,
        return_dataframe: bool = True,
    ):
        """Run query for datamodel. This will generate a SA if necessary.

        Args:
            query: BigQuery SQL query.
            dataset_id: BigQuery dataset ID.
                if None it will use the default dataset_id.
            return_dataframe: Return dataframe.
                Return dataframe if True.

        Returns:
            Query result.

        Usage:

        .. code:: python


            from pycarol import Carol
            from pycarol.bigquery import BQ

            bq = BQ(Carol())
            query = 'select * from invoice limit 10'	
            df = bq.query(query, return_dataframe=True)

        """

        self.service_account = self.get_credential()
        self.client = self._generate_client()

        dataset_id = dataset_id or self.dataset_id
        job_config = bigquery.QueryJobConfig(default_dataset=dataset_id)
        results = self.client.query(query, job_config=job_config)

        results = [dict(row) for row in results]
        if return_dataframe:
            return pd.DataFrame(results)
        else:
            return results


def query(
    carol: Carol,
    query_: str,
    service_account: T.Optional[T.Dict[str, str]] = None,
    dataset_id: T.Optional[str] = None,
) -> QueryJob:
    """Run query for datamodel.

    Args:
        query_: BigQuery SQL query.
        service_account: in case you have a service account for accessing BigQuery.
        dataset_id: BigQuery dataset ID.

    Returns:
        Query result.
    """
    if service_account is None:  # must call carol to get service account
        raise NotImplementedError(
            "You must pass a service_account. Not implemented.")

    query_ = _prepare_query(carol, query_)
    client = _generate_client(service_account)
    tenant_id = carol.tenant["mdmId"]
    dataset_id = dataset_id or f"labs-app-mdm-production.{tenant_id}"
    job_config = bigquery.QueryJobConfig(default_dataset=dataset_id)
    results = client.query(query_, job_config=job_config)
    return results


def _prepare_query(
    carol: Carol,
    query_: str,
) -> str:
    """Render template replacing variables (if any) with values.

    {{connector_name.staging_table}} is replaced by:
        `TENANTID.stg_CONNECTORID_STAGINGNAME`
    {{datamodel_name}} is replaced by `TENANTID.dm_MODELNAME`
    Variables must follow the '{{variable}}' pattern.

    Args:
        carol: Carol object.
        query_: BigQuery SQL query.

    Return:
        Query string with template rendered.
    """
    template_vars = _get_template_vars(query_)
    if len(template_vars) == 0:
        return query_

    connectors = Connectors(carol)
    connector_names = {name for name, _ in template_vars if name is not None}
    connector_map = {
        name: connectors.get_by_name(name)["mdmId"] for name in connector_names
    }

    staging_vars = filter(
        lambda conn_name: conn_name[0] is not None, template_vars)
    model_vars = filter(lambda conn_name: conn_name[0] is None, template_vars)

    replace_map = {}
    for connector_name, table_name in staging_vars:
        key = f"{connector_name}.{table_name}"
        connector_id = connector_map[connector_name]
        replace_map[key] = f"`stg_{connector_id}_{table_name}`"
    for _, table_name in model_vars:
        replace_map[table_name] = f"`dm_{table_name}`"

    def _replace_func(match) -> str:
        if match.group(2) in replace_map:
            return replace_map[match.group(2)]
        raise ValueError()

    return re.sub(r"({{\s*([0-9A-z\.]+)\s*}})", _replace_func, query_)


def _generate_client(service_account: T.Dict[str, str]) -> bigquery.Client:
    """Generate client from credentials."""
    credentials = Credentials.from_service_account_info(service_account)
    project = service_account["project_id"]
    return bigquery.Client(project=project, credentials=credentials)


REGEX = re.compile(
    r"{{\s*((?P<connector_name>[0-9A-z]+)(\.))?(?P<table_name>[0-9A-z]+.)\s*}}"
)


def _get_template_vars(query_: str) -> T.Set[T.Tuple[str, str]]:
    """Get all variables in the template.

    Variables follow the '{{connector_name.table_name}}' pattern. Connector name is
    optional.

    Args:
        query_: BigQuery SQL query.

    Return:
        Set with connector name (None when there is none) and staging/model name.
    """
    return {
        (match.group("connector_name"), match.group("table_name"))
        for match in REGEX.finditer(query_)
    }



from pytest import mark
from . import get_bytecode_tree, get_function_hash
from pycarol.pipeline.utils.hash_versioning import get_name_of_CALL_FUNCTION


def a(x):
    return x + 5


def b(z):
    return z + 5


def c(x):
    return x - 5


def d(x):
    return x + 4


def a_of_a(x):
    return a(x)


def a_of_b(x):
    return b(x)


def eternal_recursion_a(x):
    return eternal_recursion_a(x)


def eternal_recursion_b(x):
    return eternal_recursion_b(x)


def mutual_recursion_a(x):
    return mutual_recursion_b(x)


def mutual_recursion_b(x):
    return mutual_recursion_a(x)


def initial_value_a(x=3):
    return x + 5


def initial_value_b(x=5):
    return x + 5


def outer_initial_value_a():
    return initial_value_a()


def outer_initial_value_b():
    return initial_value_b()


def nested_a():
    def _nested():
        return (0)

    return _nested()


def nested_b():
    def _nested():
        return (100)

    return _nested()


def nested_c():
    def _nested():
        return (100)

    return _nested()


def nested_d():
    def _nested(x=5, y=80):
        print('kw')
        return (100 + x + y)

    return _nested(y=0, x=7)

def variable_args_a(*args):
    return (args)


def variable_args_b(*args):
    return (args)


def variable_args_c(*args):
    return args[::-1]


def builtin_a():
    print('hey')


def builtin_b():
    print('hey')


def builtin_c():
    print('hi')


def builtin_d():
    print('hey', 'hi')


def dummy_function():
    pass


def call_kwargs_a():
    return dummy_function(0, p1=10, p2=20)


def call_kwargs_b():
    return dummy_function(0, p1=10, p2=20)


def call_kwargs_c():
    return dummy_function(0, p1=10, p2=50)


def call_kwargs_d():
    return dummy_function(0, p1=10, p3=20)


def call_ex_a(x):
    return dummy_function(*x)


def call_ex_b(x):
    return dummy_function(*x, *x)


def call_ex_c(x):
    return dummy_function(**x)


def call_ex_d(x):
    return dummy_function(**x, **x)


def call_ex_e(x):
    import importlib
    d = importlib.import_module('dis')
    return dummy_function(d, *x)


def call_ex_f(x):
    return dummy_function(0, *x, **x)


def call_ex_g(x):
    return dummy_function(0, dummy_function, **x)


def call_ex_h(x, y):
    return dummy_function(0, 1, 2, [0, 1, 2], ('a', 0), *x, *x, **y, **y, p1={0, 1, 2})

import pandas as pd
import pandas
def external_import_a(x):
    return pandas.Series.cumsum(x)

def external_import_b(x):
    return pandas.Series.sum(x)

def internal_import_a(x):
    import pandas as pd
    return pd.Series(x)

def internal_import_b(x):
    import pandas as pd
    return pd.DataFrame(x)

def pick_import_a(x):
    from pandas import Series
    return Series(x)

def pick_import_b(x):
    from pandas import DataFrame
    return DataFrame(x)

from numpy.fft import fft as f1
from numpy.fft import fft2 as f2
def pick_import_c(x):
    return f1(x)

def pick_import_d(x):
    return f2(x)

from pandas import Series
def pick_import_e(x):
    return Series.cumsum(x)

def pick_import_f(x):
    return Series.sum(x)

equal_functions_list = [
    (a, a),
    (b, a),
    (a, b),
    (a_of_a, a_of_b),
    (eternal_recursion_a, eternal_recursion_b),
    (mutual_recursion_a, mutual_recursion_b),
    (variable_args_a, variable_args_b),
    (nested_b, nested_c),
    (builtin_a, builtin_b),
    (call_kwargs_a, call_kwargs_b)
]

different_functions_list = [
    (a, c),
    (b, c),
    (a, d),
    (mutual_recursion_a, eternal_recursion_a),
    (initial_value_a, a),
    (initial_value_a, initial_value_b),
    (outer_initial_value_a, outer_initial_value_b),
    (nested_a, nested_b),
    (variable_args_a, variable_args_c),
    (builtin_a, builtin_c),
    (builtin_a, builtin_d),
    (call_kwargs_a, call_kwargs_c),
    (call_kwargs_a, call_kwargs_d),
    (external_import_a,external_import_b),
    (pick_import_a,pick_import_b),
    (pick_import_c, pick_import_d),
    (pick_import_e, pick_import_f),
]

calling_functions_list = [
    call_kwargs_a,
    call_kwargs_b,
    call_kwargs_c,
    call_kwargs_d,
    call_ex_a,
    call_ex_b,
    call_ex_c,
    call_ex_d,
    call_ex_e,
    call_ex_f,
    call_ex_g,
    call_ex_h,
]

@mark.parametrize("func", calling_functions_list)
def test_find_called_function(func):
    import dis
    print(dis.dis(func))
    instructions = list(dis.get_instructions(func))
    ix = len(instructions) - 2
    inst = instructions[ix]
    # assert that the func return another function
    assert "CALL_FUNCTION" in inst.opname
    assert get_name_of_CALL_FUNCTION(ix, inst, instructions,func
                                     ) == "dummy_function"


@mark.parametrize("f1,f2", equal_functions_list)
def test_equal_functions(f1, f2):
    assert get_bytecode_tree(f1) == get_bytecode_tree(f2)


@mark.parametrize("f1,f2", different_functions_list)
def test_different_functions_robust(f1, f2):
    assert get_bytecode_tree(f1,ignore_not_implemented=True) != \
           get_bytecode_tree(f2,ignore_not_implemented=True)



all_functions = set()
for a,b in equal_functions_list + different_functions_list:
    all_functions.add(a)
    all_functions.add(b)
all_functions = [f for f in all_functions]

@mark.parametrize("f",all_functions)
def test_generate_bytecode(f):
    assert get_bytecode_tree(f,ignore_not_implemented=True)


@mark.parametrize("f",all_functions)
def test_generate_hash(f):
    print(get_bytecode_tree(f,ignore_not_implemented=True))
    assert get_function_hash(f,ignore_not_implemented=True)


from joblib import Parallel, delayed

def get_hash(f):
    name = f.__name__
    try:
        bytecode = get_bytecode_tree(f)
    except:
        bytecode = "FAIL"
    try:
        h = get_function_hash(f)
    except:
        h = "FAIL"    
    return([name, bytecode, h])


TDD_tests = False
# We place here tests that are not passing, but should not block a PR
# Typically, they are tests related to WIP
if TDD_tests:
    different_functions_list.append((internal_import_a,internal_import_b),)

    @mark.parametrize("f1,f2", different_functions_list)
    def test_different_functions(f1, f2):
        assert get_bytecode_tree(f1) != get_bytecode_tree(f2)
    @mark.parametrize("f1,f2", different_functions_list)

    def test_different_functions_robust_extended(f1, f2):
        assert get_bytecode_tree(f1,ignore_not_implemented=True) != \
            get_bytecode_tree(f2,ignore_not_implemented=True)





import builtins

def int_to_bytes(i: int) -> bytes:
    """
    Wrapper around int.to_bytes. Set some defaults parameters

    """
    return i.to_bytes(i.bit_length() // 8 + 1, 'little', signed=True)


def flat_list(tree: list) -> list:
    """
    Recursively unnest a nested list
    Args:
        tree: nested lists of unlimited depth

    Returns:
        l: flat list

    """
    l = []
    for node in tree:
        if isinstance(node, list):
            l += flat_list(node)
        else:
            l.append(node)
    return l


def build_dag(top_nodes: list, get_sons: 'function' = None) -> dict:
    """
    Extract a Direct Acyclic Graph structure of a pipeline using
    get_sons_method to fetch sons nodes

    Args:
        top_nodes: list of top tasks/nodes
        get_sons: method to extract sons node (required tasks) of a
        given node. method get_sons should return a list of nodes

    Returns:
        dag: dictionary encoding a DAG data structure.

    """
    assert isinstance(top_nodes,list)
    dag = {}

    def build_tree(task_list):
        # breadth first search
        nonlocal dag, get_sons

        # add new nodes
        for t in task_list:
            if t not in dag:
                dag[t] = get_sons(t)

        # get all nodes of this level
        sons_list = []
        for k,v in dag.items():
            for vi in v:
                if vi not in dag:
                    sons_list.append(vi)

        # recursion level wise
        if sons_list:
            build_tree(sons_list)

    build_tree(top_nodes)
    return dag


def find_root_in_dag(dag: dict) -> list:
    """
    Search in a direct acyclic graph all nodes without incoming edges
    Args:
        dag: dictionary encoding a DAG

    Returns:
        root_nodes: list of root nodes
    """
    candidates = [k for k in dag]
    for k, sons_list in dag.items():
        for it in sons_list:
            if it in candidates:
                candidates.remove(it)
    root_nodes = candidates
    return root_nodes

def find_leaf_in_dag(dag: dict) -> list:
    """
    Search in a direct acyclic graph all nodes without outcoming edges
    Args:
        dag: dictionary encoding a DAG

    Returns:
        root_nodes: list of root nodes
    """
    rev_dag = get_reverse_dag(dag)
    return find_root_in_dag(rev_dag)

def get_dag_node_level(dag: dict, ) -> dict:
    """
    Returns a dict, whose keys are nodes found in dag and values are the
    depth of the node. Root nodes have value 0.
    Args:
        dag: dictionary encoding a DAG

    Returns:
        levels: dict
    """
    # levels are initialized to 0.
    # all unreachable nodes are considered to be root nodes
    levels = {}
    for k in dag:
        levels[k] = 0

    for level, nodes_list in enumerate(breadth_first_search(dag)):
        for node in nodes_list:
            levels[node] = level

    return levels


def breadth_first_search(dag: dict, starting_nodes = [], f: 'function'=None,
                         *args,
                         **kwargs):
    """
    Generator that traverse the DAG in breadth first search mode starting in
    given nodes. At each node, execute the function f(node,*args,**kwargs).
    After executing the function, it returns a list of nodes of the current
    depth.
    Args:
        dag: dict encoding the DAG
        starting_nodes: list of starting nodes. if empty all root nodes are
        used.
        f: function to be executed at each node. it receives at least one
        argument

    Returns:
        None
    """
    if not starting_nodes:
        starting_nodes = find_root_in_dag(dag)
    yield starting_nodes

    while starting_nodes:
        if f is not None:
            # visit nodes
            for n in starting_nodes:
                f(n, *args, **kwargs)

        # get all nodes of this level
        sons_list = [n for k in starting_nodes for n in dag[k]]

        yield sons_list

        starting_nodes = sons_list


def get_reverse_dag(dag: dict) -> dict:
    """
    Returns a DAG with the same nodes as the original one, but with edges
    direction reversed. Root nodes become leaf nodes and vice-versa.
    Args:
        dag: dict encoding a DAG

    Returns:
        rev_dag: dict encoding a DAG
    """
    rev_dag = {}
    for source, v in dag.items():
        if source not in rev_dag:
            rev_dag[source] = []
        for target in v:
            if target in rev_dag:
                rev_dag[target].append(source)
            else:
                rev_dag[target] = [source]

    return rev_dag


def is_builtin(f: 'function') -> bool:
    """Returns True if f is built-in"""
    return hasattr(f, '__name__') and hasattr(builtins, f.__name__)


def is_function(f) -> bool:
    """Returns True if f is function object"""
    return hasattr(f, '__code__')


def is_code(c) -> bool:
    return hasattr(c, 'co_code')


def test_isbuiltin():
    assert is_builtin(print)
    def f():
        return None
    assert not is_builtin(f)
    from numpy.fft.fftpack import fft
    assert not is_builtin(fft)


def enumerate_with_context(instructions):
    for ix, inst in enumerate(instructions):
        yield (ix,inst,instructions)


from .task import *
from pycarol.utils.asserts import (
    assert_env_is_defined,
    dotenv_context,
)

def f1(x):
    return x+1

def f2(x):
    return x+1

def f3(x):
    return f3(x+1)

def test_empty_task():
    class Task1(Task):
        pass
    t1 = Task1()
    assert t1

def test_if_task_generates_hash():
    class Task1(Task):
        task_function = f1
    t1 = Task1()
    h = t1.hash_version()
    assert isinstance(h,int)
    assert h !=0


def test_if_tasks_has_equal_hash():
    class Task1(Task):
        task_function = f1

    class Task2(Task):
        task_function = f2

    assert Task1().hash_version() == Task2().hash_version()

def test_if_tasks_has_different_hash_1():
    class Task1(Task):
        task_function = f1
        
    class Task3(Task):
        task_function = f3

    assert Task1().hash_version() != Task3().hash_version()

def test_if_tasks_has_different_hash_2():
    class Task1(Task):
        pass
        
    class Task2(Task):
        task_function = f1

    assert Task1().hash_version() != Task2().hash_version()

if False:
    #temporarily disable possibly blocking test.
    #TODO: define carol app env in test environment

    def test_env_is_defined():
        with dotenv_context():
            assert_env_is_defined()

    def test_load_metadata():
        with dotenv_context():
            def f():
                return 34
            class Task1(Task):
                task_function = f

            params = {}
            T1 = Task1(**params)
            live_hash = T1.hash_version()
            assert live_hash is not None
            assert live_hash != 0
            T1.run()
            metadata = T1.load_metadata()
            stored_hash = metadata['hash_version']
            assert stored_hash == live_hash

    





"""
This tool helps to DataModels from json files.


"""

from ..data_models.data_model_types import DataModelTypeIds
from ..verticals import Verticals
from ..data_models.data_models import CreateDataModel

import random


class DataModelGenerator(object):

    """
    This class will create a Datamodel from a python dictionary. Today, it is not allowed nested fileds in the dictionary.

    Usage:

    .. code-block:: python

        from pycarol import PwdAuth
        from pycarol.carol import Carol
        from pycarol.utils.data_model_generator import DataModelGenerator

        json_sample = {'atende_sus': 'Sim',
                       'estabelecimento': 'Hospital Raimundo Chaar',
                       'existente': '42',
                       'tipo_do_estabelecimento': '05 - Hospital Geral',
                       'tipo_do_estabelecimento_resumo': 'Hospitais',
                       'uf': 'AC'}

        login = Carol('foo_tenant, 'foo', auth=PwdAuth('foo@totvs.com.br', 'foo123'))
        la = DataModelGenerator(login)
        la.start(json_sample, dm_name='hfair', profile_title='estabelecimento', publish=True, overwrite=True)


    """


    def __init__(self, carol):
        self.carol = carol

    def start(self, json_sample, dm_name, publish=True, profile_title=None, overwrite=False, vertical_ids=None,
              vertical_names='retail', entity_template_type_ids=None,
              entity_template_type_names='product', dm_label=None,
              group_name='Others', transaction_dm=False, ignore_field_type=False):
        """

        Start the process.

        Args:

            json_sample: `dict`
                Dictionary with key and values to be used as template for the data model
            dm_name: `str`
                Data model name
            publish: `bool` default `True`
                Publish the data model at the end.
            profile_title: `list` default `None`
                Carol's Profile title. It should be an field in json_sample
            overwrite: `bool` default `False`
                Overwrite if the data model already exists
            vertical_ids: `str` default `None`
                Vertical ID for the data model. either `vertical_ids`
            vertical_names: `str` default `retail`
                Vertical Name for the data model. either `vertical_ids`
            entity_template_type_ids: `str` default `None`
                Data model template type id
            entity_template_type_names: `str` default `product`
                Data model template type name
            dm_label: `str` default `None`
                Label for the data model
            group_name: `str` default `Others`
                Group name for the data model
            transaction_dm: `bool` default `False`
                Transaction Data model
            ignore_field_type: `bool` default `False`
                Ignore filed type when creating the data model. It means that any type conflict will be ignored. pycarol
                will use the one that already exists.

        Returns:
            None
        """

        if publish:
            if profile_title is None:
                raise ValueError('To publish the data model, `profile_title` has to be set. Use `publish=False`')
            if isinstance(profile_title, str):
                profile_title = [profile_title]
            elif not isinstance(profile_title, list):
                raise ValueError('`profile_title` has to be a list of values.')
            assert all([i in json_sample for i in profile_title]), "all profile title values should be in `json_sample`"

        if ((vertical_names is None) and (vertical_ids is None)):
            self.verticals_dict = Verticals(self.carol).all()
            vertical_names = random.choice(self.verticals_dict.keys())
            vertical_ids = self.verticals_dict.get(vertical_names)
        if ((entity_template_type_ids is None) and (entity_template_type_names is None)):
            self.entityTemplateTypesDict = DataModelTypeIds(self.carol).all()
            entity_template_type_names = random.choice(self.entityTemplateTypesDict.keys())
            entity_template_type_ids = self.entityTemplateTypesDict.get(entity_template_type_names)

        tDM = CreateDataModel(self.carol)
        tDM.create(dm_name, overwrite=overwrite, vertical_ids=vertical_ids, vertical_names=vertical_names,
                   entity_template_type_ids=entity_template_type_ids,
                   entity_template_type_names=entity_template_type_names,
                   label=dm_label, group_name=group_name, transaction_data_model=transaction_dm)

        dm_id = tDM.template_dict[dm_name]['mdmId']

        tDM.from_json(json_sample, profile_title=profile_title, publish=publish, dm_id=dm_id,
                      ignore_field_type=ignore_field_type, )
        print('Done!')



class PwdAuthCloner:
    def __init__(self, auth):
        self.user = auth.user
        self.password = auth.password

    def build(self):
        from .PwdAuth import PwdAuth
        return PwdAuth(self.user, self.password)



from enum import Enum

class Filter:
    """

    Class responsible for creating the json queries to be used in `pycarol.Query`

    Usage:

    .. code:: python

        from pycarol.filter import TYPE_FILTER, Filter, TERM_FILTER
        json_query = Filter.Builder()\\
            .must(TYPE_FILTER(value='medicalform' + "Golden"))\\
            .must(TERM_FILTER(key='mdmGoldenFieldAndValues.status.raw',value='pending'))\\
            .must_not(TERM_FILTER(key='mdmGoldenFieldAndValues.auditedbycarol',value=True))\\
            .build().to_json()

    This will create the following json query.

    .. code:: json

        {
          'mustList': [
            {
              'mdmFilterType': 'TYPE_FILTER',
              'mdmValue': 'medicalformGolden'
            },
            {
              'mdmFilterType': 'TERM_FILTER',
              'mdmKey': 'mdmGoldenFieldAndValues.status.raw',
              'mdmValue': 'pending'
            }
          ],
          'mustNotList': [
            {
              'mdmFilterType': 'TERM_FILTER',
              'mdmKey': 'mdmGoldenFieldAndValues.auditedbycarol',
              'mdmValue': True
            }
          ],
          'shouldList': [

          ],
          'aggregationList': [

          ],
          'minimumShouldMatch': 1
        }

    Using with Aggregations:

    .. code:: python

        from pycarol.filter import MINIMUM, MAXIMUM, TYPE_FILTER, Filter, TERM_FILTER
        json_query = Filter.Builder() \\
                            .type('datamodel') \\
                            .aggregation_list([MINIMUM(name='MINIMUM', params='mdm_key'),
                                               MAXIMUM(name='MAXIMUM', params='mdm_key')]) \\
                            .build().to_json()

    .. code:: json

        {
          'mustList': [
            {
              'mdmFilterType': 'TYPE_FILTER',
              'mdmValue': 'datamodel'
            }
          ],
          'mustNotList': [

          ],
          'shouldList': [

          ],
          'aggregationList': [
            {
              'type': 'MINIMUM',
              'name': 'MINIMUM',
              'params': 'mdm_key',
              'size': 10,
              'shardSize': 10,
              'minDocCount': 0
            },
            {
              'type': 'MAXIMUM',
              'name': 'MAXIMUM',
              'params': 'mdm_key',
              'size': 10,
              'shardSize': 10,
              'minDocCount': 0
            }
          ],
          'minimumShouldMatch': 1
        }


    """
    def __init__(self, builder):
        self.must_list = builder._must_list
        self.must_not_list = builder._must_not_list
        self.should_list = builder._should_list
        self.aggregation_list = builder._aggregation_list
        self.minimum_should_match = builder._minimum_should_match

    def to_json(self):
        json = {}
        json['mustList'] = [elt.to_json() for elt in self.must_list]
        json['mustNotList'] = [elt.to_json() for elt in self.must_not_list]
        json['shouldList'] = [elt.to_json() for elt in self.should_list]
        json['aggregationList'] = [elt.to_json() for elt in self.aggregation_list]
        json['minimumShouldMatch'] = self.minimum_should_match

        return json

    class Builder:
        def __init__(self, key_prefix=""):
            self._minimum_should_match = 1
            self._must_list = []
            self._must_not_list = []
            self._should_list = []
            self._aggregation_list = []
            self.key_prefix = key_prefix

        def type(self, value):
            self._must_list.append(TYPE_FILTER(value=value))
            return self

        def must(self, must):
            must.set_key_prefix(self.key_prefix)
            self._must_list.append(must)
            return self

        def must_list(self, must_list):
            assert isinstance(must_list, list)
            for must in must_list:
                must.set_key_prefix(self.key_prefix)
            self._must_list.extend(must_list)
            return self

        def must_not(self, must_not):
            must_not.set_key_prefix(self.key_prefix)
            self._must_not_list.append(must_not)
            return self

        def must_not_list(self, must_not_list):
            assert isinstance(must_not_list, list)
            for must_not in must_not_list:
                must_not.set_key_prefix(self.key_prefix)
            self._must_not_list.extend(must_not_list)
            return self

        def should(self, should):
            should.set_key_prefix(self.key_prefix)
            self._should_list.append(should)
            return self

        def should_list(self, should_list):
            assert isinstance(should_list, list)
            for should in should_list:
                should.set_key_prefix(self.key_prefix)
            self._should_list.extend(should_list)
            return self

        def aggregation(self, aggregation):
            self._aggregation_list.append(aggregation)
            return self

        def aggregation_list(self, aggregation_list):
            assert isinstance(aggregation_list, list)
            self._aggregation_list.extend(aggregation_list)
            return self

        def minimum_should_match(self, minimum_should_match):
            self._minimum_should_match = minimum_should_match
            return self

        def build(self):
            return Filter(self)

class FilterType:
    def __init__(self, filter_type, key = None, value = None, path = None, range_values = None, values_field = None,
                 mdm_format = None, flags = None, range_start = None, range_end = None, values_query = None):
        self.filter_type = filter_type
        self.key = key
        self.value = value
        self.path = path
        if range_values is not None:
            assert isinstance(range_values, list)
        self.range_values = range_values
        self.values_field = values_field
        self.mdm_format = mdm_format
        self.flags = flags
        self.range_start = range_start
        self.range_end = range_end
        if values_query is not None:
            assert values_field is not None
            assert isinstance(values_query, FilterType)
        self.values_query = values_query

    def set_key_prefix(self, key_prefix):
        if key_prefix:
            if self.key:
                self.key = key_prefix + '.' + self.key
            if self.values_field:
                self.values_field = key_prefix + '.' + self.values_field

    def to_json(self):
        json = {'mdmFilterType': self.filter_type.value}
        if self.key:
            json['mdmKey'] = self.key
        if self.value is not None:
            json['mdmValue'] = self.value
        if self.path:
            json['mdmPath'] = self.path
        if self.range_values:
            json['mdmRangeValues'] = self.range_values
        if self.values_field:
            json['mdmValuesField'] = self.values_field
        if self.mdm_format:
            json['mdmFormat'] = self.mdm_format
        if self.flags is not None:
            json['mdmFlags'] = self.flags
        if self.range_start is not None:
            json['mdmRangeStart'] = self.range_start
        if self.range_end is not None:
            json['mdmRangeEnd'] = self.range_end
        if self.values_query is not None:
            json['mdmValuesQuery'] = self.values_query.to_json()

        return json

class TYPE_FILTER(FilterType):
    def __init__(self, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.TYPE_FILTER, value=value, values_field=values_field, values_query=values_query)

class BOOL_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.BOOL_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class TERM_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.TERM_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class TERMS_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.TERMS_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class RANGE_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.RANGE_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class MATCH_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.MATCH_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class MATCH_ALL_TERMS_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.MATCH_ALL_TERMS_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class MATCH_ANY_TERM_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.MATCH_ALL_TERMS_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class TERM_FUZZY_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.TERM_FUZZY_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class TERMS_FUZZY_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.TERMS_FUZZY_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class WILDCARD_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.WILDCARD_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class WILDCARD_CUSTOM_FILTER(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.WILDCARD_CUSTOM_FILTER, key=key, value=value, values_field=values_field, values_query=values_query)

class EXISTS_FILTER(FilterType):
    def __init__(self, key, values_field = None, values_query = None):
        super().__init__(filter_type = FT.EXISTS_FILTER, key=key, values_field=values_field, values_query=values_query)

class SIMPLE_QUERY_STRING(FilterType):
    def __init__(self, key, value, values_field = None, values_query = None):
        super().__init__(filter_type = FT.SIMPLE_QUERY_STRING, key=key, value=value, values_field=values_field, values_query=values_query)

class MISSING_FILTER(FilterType):
    def __init__(self, key, values_field = None, values_query = None):
        super().__init__(filter_type = FT.MISSING_FILTER, key=key, values_field=values_field, values_query=values_query)

class GEODISTANCE_FILTER(FilterType):
    def __init__(self, path, key, range_values, values_field = None, values_query = None):
        super().__init__(filter_type = FT.GEODISTANCE_FILTER, path=path, key=key, range_values=range_values, values_field=values_field, values_query=values_query)

class NESTED(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_TERM_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_TERM_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_TERMS_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_TERMS_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_RANGE_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_RANGE_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_MATCH_ALL_TERMS_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_MATCH_ALL_TERMS_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_MATCH_ANY_TERM_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_MATCH_ANY_TERM_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_TERM_FUZZY_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_TERM_FUZZY_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_TERMS_FUZZY_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_TERMS_FUZZY_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_WILDCARD_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_WILDCARD_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_WILDCARD_CUSTOM_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_WILDCARD_CUSTOM_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_EXISTS_FILTER(FilterType):
    def __init__(self, key, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_EXISTS_FILTER, key=key, path=path, values_field=values_field, values_query=values_query)

class NESTED_SIMPLE_QUERY_STRING(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_SIMPLE_QUERY_STRING, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_MISSING_FILTER(FilterType):
    def __init__(self, key, value, path, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_MISSING_FILTER, key=key, value=value, path=path, values_field=values_field, values_query=values_query)

class NESTED_GEODISTANCE_FILTER(FilterType):
    def __init__(self, path, key, range_values, values_field = None, values_query = None):
        super().__init__(filter_type = FT.NESTED_GEODISTANCE_FILTER, path=path, key=key, range_values=range_values, values_field=values_field, values_query=values_query)

class Aggregation:
    def __init__(self, agg_type, name, params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by = None, sort_order = None, query_param = None):
        self.agg_type = agg_type
        self.name = name
        self.params = params
        if sub_aggregations is not None:
            assert isinstance(sub_aggregations,list), 'sub_aggregations must be a list'
        self.sub_aggregations = sub_aggregations
        self.size = size
        self.shard_size = shard_size
        self.min_doc_count = min_doc_count
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.query_param = query_param

    def to_json(self):
        json = {}
        json['type'] = self.agg_type.value
        json['name'] = self.name
        if self.params:
            json['params'] = self.params
        if self.size:
            json['size'] = self.size
        if self.shard_size:
            json['shardSize'] = self.shard_size
        if self.min_doc_count is not None:
            json['minDocCount'] = self.min_doc_count
        if self.sort_by:
            json['sortBy'] = self.sort_by
        if self.sort_order:
            json['sortOrder'] = self.sort_order
        if self.sub_aggregations:
            json['subAggregations'] = [agg.to_json() for agg in self.sub_aggregations]
        if self.query_param is not None:
            json['queryParam'] = self.query_param.to_json()

        return json

class TERM(Aggregation):
    def __init__(self, name='term', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.TERM, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class TERMS(Aggregation):
    def __init__(self, name='terms', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.TERMS, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class NESTED(Aggregation):
    def __init__(self, name='nested', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.NESTED, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class FILTER(Aggregation):
    def __init__(self, name='filter', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None, query_param=None):
        super().__init__(agg_type= AGG.FILTER, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order, query_param=query_param)

class MINIMUM(Aggregation):
    def __init__(self, name='minimum', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.MINIMUM, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class MAXIMUM(Aggregation):
    def __init__(self, name='maximum', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.MAXIMUM, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class AVERAGE(Aggregation):
    def __init__(self, name='average', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.AVERAGE, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class COUNT(Aggregation):
    def __init__(self, name='count', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.COUNT, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class SUM(Aggregation):
    def __init__(self, name='sum', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.SUM, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class STATS(Aggregation):
    def __init__(self, name='stats', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.STATS, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class EXTENDED_STATS(Aggregation):
    def __init__(self, name='extendedStats', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.EXTENDED_STATS, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class DATE_HISTOGRAM(Aggregation):
    def __init__(self, name='dateHistogram', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.DATE_HISTOGRAM, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class HISTOGRAM(Aggregation):
    def __init__(self, name='histogram', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.HISTOGRAM, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class RANGE(Aggregation):
    def __init__(self, name='range', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.RANGE, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class CARDINALITY(Aggregation):
    def __init__(self, name='cardinality', params=None, sub_aggregations=None, size=10, shard_size=10, min_doc_count=0, sort_by=None, sort_order=None):
        super().__init__(agg_type= AGG.CARDINALITY, name=name, params=params, sub_aggregations=sub_aggregations, size=size, shard_size=shard_size, min_doc_count=min_doc_count, sort_by=sort_by, sort_order=sort_order)

class AGG(Enum):
    TERM = "TERM"
    TERMS = "TERMS"
    NESTED = "NESTED"
    FILTER = "FILTER"
    MINIMUM = "MINIMUM"
    MAXIMUM = "MAXIMUM"
    AVERAGE = "AVERAGE"
    COUNT = "COUNT"
    SUM = "SUM"
    STATS = "STATS"
    EXTENDED_STATS = "EXTENDED_STATS"
    DATE_HISTOGRAM = "DATE_HISTOGRAM"
    HISTOGRAM = "HISTOGRAM"
    RANGE = "RANGE"
    CARDINALITY = "CARDINALITY"


class FT(Enum):
    TERM_FILTER = "TERM_FILTER"
    BOOL_FILTER = "BOOL_FILTER"
    TERMS_FILTER = "TERMS_FILTER"
    MATCH_ALL_FILTER = "MATCH_ALL_FILTER"
    MATCH_FILTER = "MATCH_FILTER"
    MATCH_ALL_TERMS_FILTER = "MATCH_ALL_TERMS_FILTER"
    MATCH_ANY_TERM_FILTER = "MATCH_ANY_TERM_FILTER"
    TERM_FUZZY_FILTER = "TERM_FUZZY_FILTER"
    TERMS_FUZZY_FILTER = "TERMS_FUZZY_FILTER"
    TYPE_FILTER = "TYPE_FILTER"
    RANGE_FILTER = "RANGE_FILTER"
    WILDCARD_FILTER = "WILDCARD_FILTER"
    WILDCARD_CUSTOM_FILTER = "WILDCARD_CUSTOM_FILTER"
    EXISTS_FILTER = "EXISTS_FILTER"
    SIMPLE_QUERY_STRING = "SIMPLE_QUERY_STRING"
    MISSING_FILTER = "MISSING_FILTER"
    GEODISTANCE_FILTER = "GEODISTANCE_FILTER"
    NESTED = "NESTED"
    NESTED_MATCH_ALL_TERMS_FILTER = "NESTED_MATCH_ALL_TERMS_FILTER"
    NESTED_MATCH_ANY_TERM_FILTER = "NESTED_MATCH_ANY_TERM_FILTER"
    NESTED_TERM_FILTER = "NESTED_TERM_FILTER"
    NESTED_TERMS_FILTER = "NESTED_TERMS_FILTER"
    NESTED_MATCH_ALL_FILTER = "NESTED_MATCH_ALL_FILTER"
    NESTED_MATCH_FILTER = "NESTED_MATCH_FILTER"
    NESTED_TERM_FUZZY_FILTER = "NESTED_TERM_FUZZY_FILTER"
    NESTED_TERMS_FUZZY_FILTER = "NESTED_TERMS_FUZZY_FILTER"
    NESTED_RANGE_FILTER = "NESTED_RANGE_FILTER"
    NESTED_WILDCARD_FILTER = "NESTED_WILDCARD_FILTER"
    NESTED_WILDCARD_CUSTOM_FILTER = "NESTED_WILDCARD_CUSTOM_FILTER"
    NESTED_SIMPLE_QUERY_STRING = "NESTED_SIMPLE_QUERY_STRING"
    NESTED_GEODISTANCE_FILTER = "NESTED_GEODISTANCE_FILTER"
    NESTED_EXISTS_FILTER = "NESTED_EXISTS_FILTER"
    NESTED_MISSING_FILTER = "NESTED_MISSING_FILTER"



from pycarol.pipeline.tools import *
from pycarol.pipeline.tools.pipeline_example import T1,T2,T3


def test_create_pipe():
    params = {}
    pipe = Pipe([T3],params)
    assert pipe.dag


def test_run_pipe():
    params = {}
    pipe = Pipe([T3],params)
    pipe.remove_all()
    pipe.run()
    assert T1(**params).output().exists()
    assert T2(**params).output().exists()
    assert T3(**params).output().exists()

def test_remove_all():
    params = {}
    pipe = Pipe([T3],params)
    pipe.remove_all()
    pipe.run()
    assert T1(**params).output().exists()
    assert T2(**params).output().exists()
    assert T3(**params).output().exists()
    pipe.remove_all()
    assert not T1(**params).output().exists()
    assert not T2(**params).output().exists()
    assert not T3(**params).output().exists()

def test_remove_upstream():
    params = {}
    pipe = Pipe([T3],params)
    pipe.remove_all()
    pipe.run()
    assert T1(**params).output().exists()
    assert T2(**params).output().exists()
    assert T3(**params).output().exists()
    pipe.remove_upstream([T2(**params)])
    assert T1(**params).output().exists()
    assert not T2(**params).output().exists()
    assert not T3(**params).output().exists()

def test_remove_orphans():
    params = {}
    pipe = Pipe([T3],params)
    pipe.remove_all()
    pipe.run()
    assert T1(**params).output().exists()
    assert T2(**params).output().exists()
    assert T3(**params).output().exists()
    T1(**params).remove()
    pipe.remove_orphans()
    assert not T1(**params).output().exists()
    assert T2(**params).output().exists()
    assert not T3(**params).output().exists()

def test_get_task_by_id():
    params = {}
    pipe = Pipe([T3],params)
    t3_id = T3(**params).task_id
    assert pipe.get_task_by_id(t3_id) == T3(**params)

def test_get_task_by_id_raises():
    params = {}
    pipe = Pipe([T3],params)
    try:
        pipe.get_task_by_id("wrog_name")
    except KeyError:
        return # success
    raise Exception("Key error waas not triggered")


# def test_remove_obsolete():
#     # need to implement metadata in targets
# test hard to automate
def test_tasks_are_instance():
    from ._tools import _tasks_are_instance
    assert not _tasks_are_instance([T1,T2])
    params = {}
    assert _tasks_are_instance([T1(**params),T2(**params)])

def test_tasks_are_class():
    from ._tools import _tasks_are_class
    assert _tasks_are_class([T1,T2])
    params = {}
    assert not _tasks_are_class([T1(**params),T2(**params)])


class Compute:
    """
    Compute.

    """

    def __init__(self, carol):
        self.carol = carol

    def get_machine_types(self):
        response = self.carol.call_api(path=f'v1/compute/machineTypes',
                                       method='GET')

        return response

    def get_app_logs(self, app_name=None, app_type='carolApp', severity='INFO', page_token=None,  filters=None,
                     page_size=10, ):
        """
        Get execution logs for an app.

        Args:
            app_name: `str` default `None`
                Carol app name. It will overwrite the app name used in Carol() initialization.
            app_type: `str` default `carolApp`
                Possible values `carolApp` or `carolConnect`
            severity: `str` default `INFO`
                Minimal log to fetch.
            page_token: `str` default `None`
                Pagination token for logs.
            filters: `list` default `None`
                List with extra filters in string format. e.g,.
                filters = ['''"insertId"="n7qvjkx9se0erdgi1"''', "OR", '''"severity"="ERROR"''']
            page_size: `int` default 10
                Number of records to return.
        Returns:

        """

        if app_name is None:
            app_name = self.carol.app_name

        filters = filters if filters else []

        filters = '\n'.join(filters)

        params = {
            "appType": app_type,
            "appName": app_name,
            "pageSize": page_size,
            "severity": severity
        }

        if page_token is None:
            url = f'v1/log/retrieveLogs'
        else:
            url = f'v1/log/retrieveLogs/nextPage'
            params.update({"pageToken": page_token})

        response = self.carol.call_api(path=url, data=filters, backoff_factor=1,
                                       status_forcelist=[502, 429, 524, 408, 504, 598, 520, 503],
                                       method='POST', params=params, content_type='text/plain')

        return response



import warnings
import json
from ..utils.importers import _import_dask, _import_pandas
from ..storage import Storage
from ..utils.miscellaneous import drop_duplicated_parquet, drop_duplicated_parquet_dask
from ..utils.deprecation_msgs import _deprecation_msgs

class DataModelView:

    def __init__(self, carol):
        self.carol = carol

    def get_by_name(self, view_name):
        return self._get(view_name, by='name')

    def get_by_id(self, view_id):
        return self._get(view_id, by='id')


    def _get(self, id,  by='name'):

        if by == 'name':
            url = f"v1/relationshipView/name/{id}"
        elif by == 'id':
            url = f"/api/v1/relationshipView/{id}"
        else:
            raise print('Type incorrect, it should be "id" or "name"')

        return self.carol.call_api(url, method='GET')

    def _build_query_params(self):
        if self.sort_by is None:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order}
        else:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order,
                                 "sortBy": self.sort_by}

    def reprocess(self, view_name=None, view_id=None, cds=False):
        """
        Reprocess the view records.

        Args:
            view_name: `str`, default `None`
                View name
            view_id: `str`, default `None`
                View ID
            cds: `bool`, default `False`
                Save view on CDS.

        Returns: `dict`
            Task response.

        """

        if view_name is not None:
            view_id = self.get_by_name(view_name=view_name)['mdmId']
        else:
            assert view_id is not None, "'view_id' or 'view_name' must be set"

        url_filter = "v1/goldenRecordView/reprocess"

        query_params = {"relationshipViewId" : view_id,
                        "cds" : cds}
        return self.carol.call_api(url_filter, params=query_params,
                                   method='POST')

    def get_all(self, offset=0, page_size=-1, sort_order='ASC',
                sort_by=None, print_status=False,
                save_file=None):


        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self._build_query_params()

        self.template_dict = {}
        self.template_data = []
        count = self.offset

        set_param = True
        self.total_hits = float("inf")
        if save_file:
            assert isinstance(save_file, str)
            file = open(save_file, 'w', encoding='utf8')
        while count < self.total_hits:
            url_filter = "v1/relationshipView"
            query = self.carol.call_api(url_filter, params=self.query_params, method='GET')

            if query['count'] == 0:
                print('There are no more results.')
                print('Expecting {}, reponse = {}'.format(self.total_hits, count))
                break
            count += query['count']
            if set_param:
                self.total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            self.template_data.extend(query)

            self.template_dict.update({i['mdmName']: {'mdmId': i['mdmId'],
                                                      'mdmRunningState': i['mdmRunningState'],
                                                      'mdmEntityType': i['mdmEntityType']}
                                       for i in query})

            self.query_params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, self.total_hits), end='\r')
            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()
        return self


    def export(self, view_name=None, view_id=None, sync_view=True, full_export=False,
               delete_previous=False, export_format='PARQUET'):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Export datamodel to s3

        This method will trigger or pause the export of the data in the datamodel to
        s3

        :param view_name: `str`, default `None`
            View Name
        :param view_id: `str`, default `None`
            View id
        :param sync_view: `bool`, default `True`
            Sync the view
        :param full_export: `bool`, default `True`
            Do a resync of the view
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :param export_format: `str`, default `PARQUET`
            Format of the data to be exported. Possible values are:
                "PARQUET", "CSV", "JSON", "EXCEL"
        :return: None
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None



    def fetch_parquet(self, view_name, merge_records=True, backend='pandas', return_dask_graph=False,
                      columns=None, return_metadata=False, callback=None, max_hits=None,
                      cds=False, max_workers=None, return_callback_result=False):

        """

        Args:
            view_name: `str`
                View name to be imported
            merge_records: `bool`, default `True`
                This will keep only the most recent record exported. Sometimes there are updates and/or deletions and
                one should keep only the last records.
            backend: ['dask','pandas'], default `dask`
                if to use either dask or pandas to fetch the data
            return_dask_graph: `bool`, default `false`
                If to return the dask graph or the dataframe.
            columns: `list`, default `None`
                List of columns to fetch.
            return_metadata: `bool`, default `False`
                To return or not the fields ['mdmId', 'mdmCounterForEntity']
            callback: `callable`, default `None`
                Function to be called each downloaded file.
            max_hits: `int`, default `None`
                Number of records to get.
            cds: `bool`, default `False`
                    Get view data from CDS.
            max_workers: `int` default `None`
                Number of workers to use when downloading parquet files with pandas back-end.
            return_callback_result `bool` default `False`
                If a callback is used, it will return the result of the response of the callback. This will skip all the
                operation to merge records and return selected columns.
        :return:
        """

        if callback:
            assert callable(callback), \
                f'"{callback}" is a {type(callback)} and is not callable.'

        if isinstance(columns, str):
            columns = [columns]

        assert backend == 'dask' or backend == 'pandas'

        if return_dask_graph:
            assert backend == 'dask'

        if not cds:
            _deprecation_msgs("`cds` option will be removed from pycarol 3.33. Consider use `cds=True`"
                              " to avoid problems. ")

        if columns:
            columns.extend(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated'])

        import_type = 'view_cds'
        storage = Storage(self.carol)
        token_carolina = storage.backend.carolina.token
        storage_space = storage.backend.carolina.get_bucket_name(import_type)

        if backend == 'dask':
            d = _import_dask(storage=storage, view_name=view_name, import_type=import_type,
                             merge_records=merge_records, return_dask_graph=return_dask_graph,
                             columns=columns, )

        elif backend == 'pandas':

            d = _import_pandas(storage=storage, view_name=view_name, golden=True, columns=columns, callback=callback,
                               max_hits=max_hits, import_type=import_type, max_workers=max_workers,
                               token_carolina=token_carolina,
                               )
            if d is None:
                warnings.warn("No data to fetch!", UserWarning)
                return d

        else:
            raise ValueError(f'backend should be either "dask" or "pandas" you entered {backend}')

        if (return_callback_result) and (callback is not None):
            return d

        if merge_records:
            if (not return_dask_graph) or (backend == 'pandas'):
                d = drop_duplicated_parquet(d)
            else:
                d = drop_duplicated_parquet_dask(d)

        if not return_metadata:
            to_drop = set(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated']).intersection(set(d.columns))
            d = d.drop(labels=to_drop, axis=1)

        return d


    def export_all(self, sync_view=True, full_export=False, delete_previous=False):
        """

        @DEPRECATED. This function was removed in pycarol 3.34
        Export all data model view

        This method will trigger or pause the export of the data in the data model view to CDS

        :param sync_view: `bool`, default `True`
            Sync the data model view
        :param full_export: `bool`, default `True`
            Do a resync of the data model view
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :return: None
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None


    def _get_view_export_stats(self):
        """
        @DEPRECATED. This function was removed in pycarol 3.34

        Get export status for views

        :return: `dict`
            dict with the information of which data model view is exporting its data.
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None






def test_luigi_extension():
    import pycarol.luigi_extension
    import pycarol.luigi_extension.targets
    from pycarol.luigi_extension.targets import PickleTarget
    from pycarol.luigi_extension.task import Task




import codecs
from pathlib import Path
import re

from setuptools import setup, find_packages

here = Path(__file__).absolute().parent

min_requires = [
    "deprecated",
    "gcsfs>=0.3.0,<0.7",
    "google-auth",
    "google-auth-httplib2",
    "google-cloud-bigquery>=2.26.0",
    "google-cloud-core>=1.4.1",
    "google-cloud-storage",
    "python-dotenv",
    "requests",
    "retry",
    "tqdm",
    "urllib3",
]

dataframe_requires = [
    "pandas>=0.23.4,!=1.0.4",
    "numpy>=1.16.3",
    "joblib>=0.11",
    "pyarrow>=0.15.1,<1.0.0",
]
extras_require = {
    "dataframe": min_requires + dataframe_requires,
    "pipeline": min_requires + dataframe_requires + ["luigi", "papermill"],
    "onlineapp": min_requires + ["flask>=1.0.2", "redis"],
    "dask": min_requires + ["dask[complete]"],
    "dev": min_requires + ["pytest", "bumpversion", "sphinx-rtd-theme", "sphinx"],
}
extras_require["complete"] = sorted({v for req in extras_require.values() for v in req})


def read(parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(here / parts, "r") as fp:
        return fp.read()


readme_note = """\
.. note::
   For the latest source, discussion, etc, please visit the
   `GitHub repository <https://github.com/totvslabs/pyCarol>`_\n\n
"""

with open("README.rst") as fobj:
    long_description = readme_note + fobj.read()


def find_version(file_paths):
    version_file = read(file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


setup(
    name="pycarol",
    setup_requires=["wheel"],
    packages=find_packages(exclude=["docs", "doc"]),
    version=find_version("pycarol/__init__.py"),
    license="TOTVS",
    description="Carol Python API and Tools",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="TotvsLabs",
    maintainer="TOTVS Labs",
    author_email="ops@totvslabs.com",
    url="https://github.com/totvslabs/pyCarol",
    keywords=["Totvs", "Carol.ai", "AI"],
    install_requires=min_requires,
    extras_require=extras_require,
    classifiers=[
        # Chose either "3 - Alpha",
        # "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Development Status :: 5 - Production/Stable",
        # Define that your audience are developers
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
)



from .carol import Carol
from .data_models import DataModel
from .staging import Staging
from .apps import Apps
from .cds import CDSGolden, CDSStaging
from .connectors import Connectors
from .query import Query
from .storage import Storage
from .carolina import Carolina
from .tasks import Tasks
from .subscription import Subscription
from functools import partial

from .subscription import Subscription


CAROL_KNOWN_MODULES = {
    'datamodel': DataModel,
    'staging': Staging,
    'apps': Apps,
    'cds_golden': CDSGolden,
    'cds_staging': CDSStaging,
    'query': Query,
    # 'storage': Storage,  #Had to remove these two because they have google client that is unpickleable. This breaks if trying to use in any multiprocessing lib.
    # 'carolina': Carolina,
    'task': Tasks,
    'subscription': Subscription,
    'connector': Connectors,
}


class CarolAPI(Carol):
    """Encapsulates some pycarol modules

    Args:

        domain: `str`. default `None`.
            Tenant name. e.x., domain.carol.ai
        app_name: `str`. default `None`.
            Carol app name.
        auth: `PwdAuth` or `ApiKeyAuth`.
            object Auth Carol object to handle authentication
        connector_id: `str` , default `__CONNECTOR_PYCAROL__`.
            Connector Id
        port: `int` , default 443.
            Port to be used (when running locally it could change)
        verbose: `bool` , default `False`.
            If True will print the header, method and URL of each API call.
        organization: `str` , default `None`.
            Organization domain.
        environment: `str`, default `carol.ai`,
            Which Carol's environment to use. There are three possible values today.

                1. 'carol.ai' for the production environment
                2. 'karol.ai' for the explore environment
                3. 'qarol.ai' for the QA environment

        host: `str` default `None`
            This will overwrite the host used. Today the host is:

                1. if organization is None, host={domain}.{environment}
                2. else host={organization}.{environment}

            See Carol._set_host.

         user:  `str` default `None`
            User
         password: `str` default `None`
            User passowrd
         api_key:  `str` default `None`
            Carol's Api Key

    OBS:
        In case all parameters are `None`, pycarol will try yo find their values in the environment variables.
        The values are:

             1. `CAROLTENANT` for domain
             2. `CAROLAPPNAME` for app_name
             3. `CAROLAPPOAUTH` for auth
             4. `CAROLORGANIZATION` for organization
             5. `CAROLCONNECTORID` for connector_id
             6. `CAROL_DOMAIN` for environment
             7. `CAROLUSER` for carol user email
             8. `CAROLPWD` for user password.

    Usage:

    .. code:: python

            from pycarol import CarolAPI
            carol = CarolAPI()
            # saving a file using storage
            carol.storage.save(name='myfile.csv', obj='/local/file/.csv',  format='file')
            #fetch data from staging
            df = carol.staging.fetch_parquet(staging_name=staging,  
                                        connector_name=connector_name,)
            # show all the available modules. 
            carol._all_modules

    If one wants to use a module not in "carol._all_modules" it is possible to add new modules using:

    .. code:: python

            from pycarol import CarolAPI
            carol = CarolAPI()
            carol.add_module('new_module', 'ANewModule')

    Note that this module will receive pycarol.Carol as the first argument. If the module require extra arguments,
    use:

    .. code:: python

        from pycarol import CarolAPI
        carol = CarolAPI()
        carol.add_module('new_module', 'ANewModule', allow_args=True)

        carol.new_module(extra_arg, new_kwarg='foo').do_something()

    """

    def __init__(self, domain=None, app_name=None, auth=None, connector_id=None, port=443, verbose=False,
                 organization=None, environment=None, host=None, user=None, password=None, api_key=None):

        super(CarolAPI, self).__init__(domain=domain, app_name=app_name, auth=auth, connector_id=connector_id, port=port, verbose=verbose,
                                       organization=organization, environment=environment, host=host, user=user, password=password, api_key=api_key)

        self._all_modules = set()
        self._create_context()

    def _create_context(self):

        for module_name, module in CAROL_KNOWN_MODULES.items():

            if module is Query:
                allow_args = True
            else:
                allow_args = False

            self.add_module(module_name, module, allow_args=allow_args)

    def add_module(self, module_name, module, allow_args=False):
        """Adds module to context
        Args:
            module_name (str): attribute to be created
            module (pyCarol modude): the module will be inicialized with the pycarol.Carol instance.
            allow_args (bool): use functools.partial to allow passing arguments when calling the module. 
        """
        if allow_args:
            module = partial(module, self)
            setattr(self, module_name, module)
        else:
            setattr(self, module_name, module(self))
        self._all_modules.update([module_name])



import json
import warnings
import asyncio

from .schema_generator import carolSchemaGenerator
from .connectors import Connectors
from .storage import Storage
from .utils.importers import _import_dask, _import_pandas
from .utils import async_helpers
from .utils.miscellaneous import stream_data
from . import _CAROL_METADATA_STAGING, _NEEDED_FOR_MERGE, _CAROL_METADATA_UNTIE_STAGING
from .utils.miscellaneous import drop_duplicated_parquet, drop_duplicated_parquet_dask
from .utils.deprecation_msgs import _deprecation_msgs

_SCHEMA_TYPES_MAPPING = {
    "geopoint": str,
    "long": int,
    "double": float,
    "nested": str,
    "string": str,
    "base64": str,
    "date": str,
    "boolean": bool
}


class Staging:

    """

    Class to send data to Carol.

    """

    def __init__(self, carol):
        self.carol = carol

    def send_data(self, staging_name, data=None, connector_name=None, connector_id=None, step_size=500,
                  print_stats=True, gzip=True, auto_create_schema=False, crosswalk_auto_create=None,
                  flexible_schema=False, force=False,  max_workers=2,  dm_to_delete=None,
                  async_send=False, carol_data_storage=False, storage_only=True, carol_sync=False):
        """
        Send data to a staging table in Carol.

        Args:
            staging_name:  `str`,
                Staging name to send the data.
            data: pandas data frame, json. default `None`
                Data to be send to Carol
            connector_name: `str`, default `None`
                Connector name where the staging should be. Either `connector_name` or `connector_id` need to be set.
            connector_id: `str`, default `None`
                Connector Id where the staging should be. Either `connector_name` or `connector_id` need to be set.
            step_size: `int`, default `500`
                Number of records to be sent in each iteration. Max size for each batch is 10MB.
            print_stats:`bool`, default `True`
                If print the number of records sent
            gzip:`bool`, default `True`
                If send each batch as a gzip file.
            auto_create_schema:`bool`, default `False`
                If to auto create the schema for the data being sent.
            crosswalk_auto_create: `list`, default `None`
                If `auto_create_schema=True`, crosswalk list of fields.
            flexible_schema: `bool`, default `False`
                If `auto_create_schema=True`, to use a flexible schema.
            force: `bool`, default `False`
                pycarol will check for duplicated values given the crosswalk.
                If `force=True` it will not check. If `False` it will check for duplicates and raise an error.
            max_workers: `int`, default `2`
                To be used with `async_send=True`. Number of threads to use when sending.
            dm_to_delete: `str`, default `None` DEPRECATED.
                Name of the data model to be erased before send the data.
            async_send: `bool`, default `False`
                To use async to send the data. This is much faster than a sequential send.
                It can conflict with jupyter notebooks process. To run this inside a jupyter notebook, use

                .. code:: python

                    import nest_asyncio
                    nest_asyncio.apply()

            carol_data_storage: `bool`, default `False`
                Deprecated, Use `storage_only`
            storage_only: `bool`, default `False`
                Send data only to CDS.
            carol_sync: `bool`, default `False`
                Send and wait data to be processed in Carol

        """

        if dm_to_delete is not None:
            _deprecation_msgs("`dm_to_delete` is deprecated and has no action.")

        if carol_data_storage:
            _deprecation_msgs("`carol_data_storage` is deprecated and has no action.")

        if not storage_only:
            _deprecation_msgs("`storage_only` will be irrelevant. All data will be send to CDS as default.")


        self.gzip = gzip
        extra_headers = {}
        content_type = 'application/json'
        if self.gzip:
            content_type = None
            extra_headers["Content-Encoding"] = "gzip"
            extra_headers['content-type'] = 'application/json'

        if connector_name:
            connector_id = self._connector_by_name(connector_name)
        else:
            if connector_id is None:
                connector_id = self.carol.connector_id

        is_df = False
        _crosswalk = None
        if isinstance(data, str):
            data = json.loads(data)
            data_size = len(data)
            _sample_json = data[0]
        elif isinstance(data, list):
            data_size = len(data)
            _sample_json = data[0]
        elif isinstance(data, dict):
            data = [data]
            data_size = len(data)
            _sample_json = data[0]
        else:
            import pandas as pd
            if isinstance(data, pd.DataFrame):
                is_df = True
                data_size = data.shape[0]
                _sample_json = data.iloc[0].to_json(date_format='iso')
            else:
                raise ValueError('`data` should be either a stringfied json, a list of dictionaries, a dictionary or a pd.DataGrame ')


        if (not isinstance(data, list)) and (not is_df):
            data = [data]
            data_size = len(data)

        if auto_create_schema:

            if crosswalk_auto_create is None:
                raise ValueError("You should provide a crosswalk. Use `crosswalk_auto_create` parameter.")

            schema = self.get_schema(staging_name, connector_id=connector_id)
            if not schema:
                overwrite = False
            else:
                overwrite = True
            self.create_schema(_sample_json, staging_name, connector_id=connector_id, export_data=carol_data_storage,
                               crosswalk_list=crosswalk_auto_create, overwrite=overwrite, mdm_flexible=flexible_schema)
            _crosswalk = crosswalk_auto_create
            print('provided crosswalk ', _crosswalk)

        if is_df and not force:
            if _crosswalk is None:
                schema = self.get_schema(staging_name, connector_id=connector_id)
                _crosswalk = schema["mdmCrosswalkTemplate"]["mdmCrossreference"].values()
                _crosswalk = list(_crosswalk)[0]
                print('fetched crosswalk ', _crosswalk)

            if data.duplicated(subset=_crosswalk).sum() >= 1:
                raise Exception("crosswalk is not unique on data frame. set force=True to send it anyway.")

        if not storage_only and not carol_sync:
            url = f'v2/staging/tables/{staging_name}&returnData=false&connectorId={connector_id}'
        elif carol_sync:
            step_size = min(100, step_size)
            url = f'v2/staging/tables/{staging_name}/sync?&connectorId={connector_id}&processMerge=true'
        else:
            url = f'v2/staging/intake/{staging_name}?returnData=false&connectorId={connector_id}'
        
        self.cont = 0
        if async_send:
            loop = asyncio.get_event_loop()
            future = asyncio.ensure_future(async_helpers.send_data_asynchronous(carol=self.carol,
                                                                                data=data,
                                                                                step_size=step_size,
                                                                                url=url,
                                                                                extra_headers=extra_headers,
                                                                                content_type=content_type,
                                                                                max_workers=max_workers,
                                                                                compress_gzip=self.gzip))
            loop.run_until_complete(future)

        else:
            for data_json, cont in stream_data(data=data,
                                               step_size=step_size,
                                               compress_gzip=self.gzip):

                self.carol.call_api(url, data=data_json, extra_headers=extra_headers, content_type=content_type,
                                    status_forcelist=[502, 429, 524, 408, 504, 598, 520, 503, 500],
                                    method_whitelist=frozenset(['POST'])
                                    )
                if print_stats:
                    print('{}/{} sent'.format(cont, data_size), end='\r')

    def get_schema(self, staging_name, connector_name=None, connector_id=None):

        """

        Return the staging table schema.

        Args:
            staging_name:  `str`,
                Staging name
            connector_name: `str`, default `None`
                Connector name. Either `connector_name` or `connector_id` need to be set.
            connector_id: `str`, default `None`
                Connector Id. Either `connector_name` or `connector_id` need to be set.

        Returns: `dict`
            Staging table schema

        """

        query_string = None
        if connector_name:
            connector_id = self._connector_by_name(connector_name)

        if connector_id:
            query_string = {"connectorId": connector_id}
        try:
            return self.carol.call_api(f'v2/staging/tables/{staging_name}/schema', method='GET',
                                       params=query_string)
        except Exception:
            return None

    def create_schema(self, fields_dict=None, staging_name=None, connector_id=None, connector_name=None,
                      mdm_flexible=False,  crosswalk_name=None, crosswalk_list=None, overwrite=False, auto_send=True,
                      export_data=False, data=None):
        """

        Args:
            fields_dict: `dict`, `list of dicts`, `pandas.DataFrame`, default `None`
                Data to create schema from. `fields_dict` will be removed in the future. Use parameter `data`
            staging_name: `str`,
                Staging name to send the data.
            connector_id: `str`, default `None`
                Connector name where the staging should be.
            connector_name: `str`, default `None`
                Connector name where the staging should be.
            mdm_flexible:  `bool`, default `False`
                If flexible schema.
            crosswalk_name: `None`, default `staging_name`
                Crosswalk name. Most of the time it should be the staging name.
            crosswalk_list: `list`, default `None`
                Crosswalk list of fields.
            overwrite: `bool`, default `False`
                if already exists, overwrite current schema
            auto_send: `bool`, default `True`
                Send the schema after creating.
            export_data: `bool`, default `False`
                Export data to CDS for this staging. This is a manual export.
            data: `json`, `list of dicts`, `pandas.DataFrame`, default `None`
                Data to create schema from.

        """
        import pandas as pd
        if export_data is not None:
            _deprecation_msgs("`export_data` is deprecated and has no action.")

        assert staging_name is not None, 'staging_name must be set.'
        assert fields_dict is not None or data is not None, 'fields_dict or df must be set'

        if fields_dict is not None:
            warnings.warn(
                "fields_dict will be deprecated, use `data`",
                DeprecationWarning, stacklevel=3
            )
            data = fields_dict

        if connector_name:
            connector_id = self._connector_by_name(connector_name)
        else:
            assert connector_id, f'connector_id or connector name should be set.'

        if data is not None:
            if isinstance(data, pd.DataFrame):
                data = data.iloc[0].to_dict()

        if isinstance(data, list):
            data = data[0]

        if isinstance(data, dict):
            schema = carolSchemaGenerator(data)
            schema = schema.to_dict(mdmStagingType=staging_name, mdmFlexible=mdm_flexible,
                                    crosswalkname=crosswalk_name, crosswalkList=crosswalk_list)
        elif isinstance(data, str):
            schema = carolSchemaGenerator.from_json(data)
            schema = schema.to_dict(mdmStagingType=staging_name, mdmFlexible=mdm_flexible,
                                    crosswalkname=crosswalk_name, crosswalkList=crosswalk_list)
        else:
            print('Behavior for type %s not defined!' % type(data))

        if auto_send:
            self.send_schema(schema=schema, staging_name=staging_name, connector_id=connector_id, overwrite=overwrite)
        else:
            return schema

    def send_schema(self, schema, staging_name=None, connector_id=None, connector_name=None,
                    overwrite=False):

        """

        Send schema to Carol

        Args:
            schema: `dict`
                Dictianary with the schema to be sent
            staging_name: `str`, default `None`
                Staging name to send the data. If empty it will get from the `schema`
            connector_id: `str`, default `None`
                Connector name. If empty it will get from the `schema`
            connector_name: `str`, default `None`
                Connector name. If empty it will get from the `schema`
            overwrite: `bool`, default `False`
                if already exists, overwrite current schema

        Returns: `None`

        """

        if connector_name:
            connector_id = self._connector_by_name(connector_name)

        if staging_name is None:
            staging_name = schema.get('mdmStagingType')
            assert staging_name is not None, f"staging_name should be given or defined in the schema."

        query_string = {"connectorId": connector_id}
        if connector_id is None:
            connector_id = self.carol.connector_id
            query_string = {"connectorId": connector_id}

        has_schema = self.get_schema(staging_name, connector_id=connector_id) is not None
        if has_schema and overwrite:
            method = 'PUT'
        else:
            method = 'POST'

        resp = self.carol.call_api('v2/staging/tables/{}/schema'.format(staging_name), data=schema, method=method,
                                   params=query_string)

    def drop_staging(self, staging_name=None, connector_id=None, connector_name=None,
                     reject_on_no_schema=False, reject_on_etl_existence=False, reject_on_mapping_existence=False
                     ):
        """

        Args:

            staging_name:  `str`,
                Staging name.
            connector_name: `str`, default `None`
                Connector name.
            connector_id: `str`, default `None`
                Connector Id.
            reject_on_no_schema: `bool` default `False`
                Do not drop if no schema found.
            reject_on_etl_existence: `bool` default `False`
                Raise an error if there exists a ETL for this staging table
            reject_on_mapping_existence: `bool` default `False`
                Raise an error if there exists a mapping for this staging table

        Returns:

            Carol response

        """

        if connector_name:
            connector_id = self._connector_by_name(connector_name)
        elif connector_id is None:
            raise ValueError("Either `connector_id` or `connector_name` should be set")

        query_string = {"connectorId" : connector_id,
                        "rejectOnNoSchema" : reject_on_no_schema,
                        "rejectOnETLExistence" : reject_on_etl_existence,
                        "rejectOnMappingExistence" : reject_on_mapping_existence}

        resp = self.carol.call_api(f'v2/staging/tables/{staging_name}/drop', method='DELETE', params=query_string)
        return resp

    def _check_crosswalk_in_data(self, schema, _sample_json):
        crosswalk = schema["mdmCrosswalkTemplate"]["mdmCrossreference"].values()
        if all(name in _sample_json for name in crosswalk):
            pass

    def _connector_by_name(self, connector_name):
        """
        Get connector id given connector name

        :param connector_name: `str`
            Connector name
        :return: `str`
            Connector Id
        """
        return Connectors(self.carol).get_by_name(connector_name)['mdmId']

    def fetch_parquet(self, staging_name, connector_id=None, connector_name=None, backend='pandas',
                      merge_records=True, return_dask_graph=False, columns=None, max_hits=None,
                      return_metadata=False, callback=None, cds=True, max_workers=None, file_pattern=None,
                      return_callback_result=False):
        """

        Fetch parquet from a staging table.

        Args:
            staging_name: `str`,
                Staging name to fetch parquet of
            connector_id: `str`, default `None`
                Connector id to fetch parquet of
            connector_name: `str`, default `None`
                Connector name to fetch parquet of
            backend: ['dask','pandas'], default `dask`
                if to use either dask or pandas to fetch the data
            merge_records: `bool`, default `True`
                This will keep only the most recent record exported. Sometimes there are updates and/or deletions and
                one should keep only the last records.
            return_dask_graph: `bool`, default `false`
                If to return the dask graph or the dataframe.
            columns: `list`, default `None`
                List of columns to fetch.
            max_hits: `int`, default `None`
                Number of records to get. This only should be user for tests.
            return_metadata: `bool`, default `False`
                To return or not the fields like ['mdmId', 'mdmCounterForEntity', etc.]
            callback: `callable`, default `None`
                Function to be called each downloaded file.
            cds: `bool`, default `True`
                Get staging data from CDS.
            max_workers: `int` default `None`
                Number of workers to use when downloading parquet files with pandas back-end.
            file_pattern: `str` default `None`
                File pattern to filter data when fetching from CDS. e.g.
                file_pattern='2019-11-25' will fetch only CDS files that start with `2019-11-25`.
            return_callback_result `bool` default `False`
                If a callback is used, it will return the result of the response of the callback. This will skip all the
                operation to merge records and return selected columns.

        Returns: `pandas.DataFrame`
            DataFrame with the staging data.

        """
        if return_metadata:
            _meta_cols = _CAROL_METADATA_STAGING
        else:
            _meta_cols = _NEEDED_FOR_MERGE

        if callback and not callable(callback):
            raise TypeError(f'"{callback}" object is not callable')

        if backend not in ('dask','pandas'):
            raise ValueError(f"Backend options are 'dask','pandas' {backend} was given")

        if return_dask_graph and backend != 'dask':
            warnings.warn('`return_dask_graph` has no use when `backend!=dask`')

        if connector_name:
            connector_id = self._connector_by_name(connector_name)
        else:
            assert connector_id

        if columns:
            mapping_columns = columns
            columns = [i.replace("-", "_") for i in columns]
        else:
            _staging = self.get_schema(staging_name=staging_name, connector_id=connector_id)
            if not _staging:
                raise ValueError(f"{staging_name} does not exist for connector ID {connector_id}")
            mapping_columns = list(_staging['mdmStagingMapping']['properties'].keys())
            columns = [i.replace("-", "_") for i in mapping_columns]

        columns.extend(_meta_cols)
        mapping_columns = dict(zip([i.replace("-", "_") for i in columns], mapping_columns))

        # validate export
        if not cds:
            _deprecation_msgs("`cds` option was removed. Returning CDS data.")

        import_type = 'staging_cds'
        storage = Storage(self.carol)
        token_carolina = storage.backend.carolina.token
        storage_space = storage.backend.carolina.get_bucket_name(import_type)

        if backend == 'dask':
            d = _import_dask(storage=storage, connector_id=connector_id, staging_name=staging_name,
                             merge_records=merge_records, import_type=import_type, return_dask_graph=return_dask_graph,
                             mapping_columns=mapping_columns,
                             columns=columns, max_hits=max_hits)

        elif backend == 'pandas':
            import pandas as pd
            d = _import_pandas(storage=storage, connector_id=connector_id, max_workers=max_workers,
                               token_carolina=token_carolina, storage_space=storage_space,
                               staging_name=staging_name, import_type=import_type,  columns=columns,
                               max_hits=max_hits, callback=callback, mapping_columns=mapping_columns,
                               file_pattern=file_pattern)

            # TODO: Do the same for dask backend
            if d is None:
                warnings.warn(f'No data to fetch! {staging_name} has no data', UserWarning)
                cols_keys = list(self.get_schema(
                    staging_name=staging_name, connector_id=connector_id
                )['mdmStagingMapping']['properties'].keys())
                cols_keys = [i.replace("-", "_") for i in cols_keys]

                if return_metadata:
                    cols_keys.extend(_meta_cols)

                elif columns:
                    columns = [i for i in columns if i not in _meta_cols]

                d = pd.DataFrame(columns=cols_keys)
                for key, value in self.get_schema(staging_name=staging_name,
                                                  connector_id=connector_id)['mdmStagingMapping'][
                    'properties'].items():
                    key = key.replace("-", "_")
                    d.loc[:, key] = d.loc[:, key].astype(_SCHEMA_TYPES_MAPPING.get(value['type'], str), copy=False)
                if columns:
                    columns = list(set(columns))
                    d = d[list(set(columns))]
                return d.rename(columns=mapping_columns)
        else:
            raise ValueError(f'backend should be either "dask" or "pandas" you entered {backend}')

        if (return_callback_result) and (callback is not None):
            return d

        if merge_records:
            if (not return_dask_graph) or (backend == 'pandas'):
                d = drop_duplicated_parquet(d, untie_field=_CAROL_METADATA_UNTIE_STAGING)
            else:
                d = drop_duplicated_parquet_dask(d, untie_field=_CAROL_METADATA_UNTIE_STAGING)

        if not return_metadata:
            to_drop = set(_meta_cols).intersection(set(d.columns))
            d = d.drop(labels=to_drop, axis=1)

        return d

    def export(self, staging_name, connector_id=None, connector_name=None, sync_staging=True, full_export=False,
               delete_previous=False):
        """
        @DEPRECATED. This function was removed in pycarol 3.34

        Export Staging from RT to CDS

        This method will trigger or pause the export of the data in the staging to
        CDS.

        Args:
            staging_name: `str`, default `None`
                Staging name to send the data. If empty it will get from the `schema`
            connector_id: `str`, default `None`
                Connector name. If empty it will get from the `schema`
            connector_name: `str`, default `None`
                Connector name. If empty it will get from the `schema`
            sync_staging: `bool` default `True`
                Start export for this staging
            full_export: `bool` default `False`
                Full export sync for this staging
            delete_previous: `bool` default `False`
                Delete previous data.

        Returns: `None`


        Usage:
        To trigger the export the first time:

        .. code:: python


            from pycarol.staging import Staging
            from pycarol.auth.PwdAuth import PwdAuth
            from pycarol.carol import Carol
            login = Carol()
            stag  = Staging(login)
            stag.export(staging, connector_name=connector_name,sync_staging=True)

            #To do a resync, that is, start the sync from the begining without delete old data
            stag.export(staging, connector_name=connector_name,sync_staging=True, full_export=True)

            #To delete the old data:
            stag.export(staging, connector_name=connector_name,sync_staging=True, full_export=True, delete_previous=True)

            #To Pause a sync:
            stag.export(staging, connector_name=connector_name,sync_staging=False)


        """
        _deprecation_msgs("This function was removed from pyCarol")

        return None

    def export_all(self, connector_id=None, connector_name=None, sync_staging=True, full_export=False,
                   delete_previous=False):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Export all Stagings from a connector to s3

        This method will trigger or pause the export of all stagings  to
        s3.

        :param sync_staging: `bool`, default `True`
            Sync the data model
        :param connector_name: `str`
            Connector name
        :param connector_id: `str`
            Connector id
        :param full_export: `bool`, default `True`
            Do a resync of the data model
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :return: None

        Usage: See `Staging.export()`
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None

    def _get_staging_export_stats(self):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Get export status for data models

        :return: `dict`
            dict with the information of which staging table is exporting its data.
        """

        _deprecation_msgs("This function was removed from pyCarol")
        return None


    def get_mapping_snapshot(self, connector_id, mapping_id, entity_space='PRODUCTION', reverse_mapping=False):

        self.snap = {}
        querystring = {"entitySpace": entity_space, "reverseMapping": reverse_mapping}

        url = f"v1/connectors/{connector_id}/entityMappings/{mapping_id}/snapshot"
        response = self.carol.call_api(url, method='GET', params=querystring, )

        mapping_name = response.get('entityMappingName')
        return {mapping_name: response}

    def delete_mapping(self, staging_name=None, connector_id=None, connector_name=None, mapping_id=None,
                       entity_space='PRODUCTION'):

        cc = Connectors(self.carol)
        st = cc.get_dm_mappings(all_connectors=True)

        if mapping_id is None:
            assert staging_name is not None, "staging_name should be set."
            if (connector_id is None) and (connector_name is not None):
                connector_id = Connectors(self.carol).get_by_name(connector_name)['mdmId']
                entity = [i['mdmId'] for i in st
                          if (i.get('mdmConnectorId') == connector_id) and
                          (i.get('mdmStagingType') == staging_name)]
                assert len(entity) == 1, f'No data model mapped for {staging_name}'
                mapping_id = entity[0]

            elif connector_id is None:
                entity = [i for i in st if (i.get('mdmStagingType') == staging_name)]

                if len(entity) > 1:
                    raise KeyError(f'There are more than one connector for staging {staging_name}')
                elif len(entity) < 1:
                    raise KeyError(f'No data model mapped for {staging_name}')
                entity = entity[0]
                connector_id = entity['mdmConnectorId']
                mapping_id = entity['mdmId']
            elif connector_id:
                entity = [i['mdmId'] for i in st
                          if (i.get('mdmConnectorId') == connector_id) and
                          (i.get('mdmStagingType') == staging_name)]

                assert len(entity) == 1, f'No mapping for {staging_name}'
                mapping_id = entity[0]

        url_mapping = f'v1/connectors/{connector_id}/entityMappings/{mapping_id}'
        querystring = {"entitySpace": entity_space, "reverseMapping": "false"}

        self.carol.call_api(url_mapping, method='DELETE', params=querystring, )

    def _check_if_exists(self, connector_id, staging_name):
        conn = Connectors(self.carol)

        try:
            mappings = conn.get_dm_mappings(connector_id=connector_id, staging_name=staging_name)
            return mappings
        except Exception as e:
            if 'Entity mapping not found' in str(e):
                return None
            else:
                raise e

    def mapping_from_snapshot(self, mapping_snapshot, connector_id=None, connector_name=None,
                              publish=True, overwrite=False):

        if connector_name:
            connector_id = self._connector_by_name(connector_name)
        else:
            assert connector_id

        staging_name = mapping_snapshot.get('stagingEntityType')
        assert staging_name, f"Snapshot incomplete, `stagingEntityType` not in snapshot"

        _mappings = self._check_if_exists(connector_id=connector_id, staging_name=staging_name)

        if (_mappings is not None) and overwrite:
            _mapping_id = [i['mdmId'] for i in _mappings]
            assert len(_mapping_id) == 1
            _mapping_id = _mapping_id[0]
            method = 'PUT'
            url_mapping = f"v1/connectors/{connector_id}/entityMappings/{_mapping_id}/snapshot"
        else:
            method = 'POST'
            url_mapping = f'v1/connectors/{connector_id}/entityMappings/snapshot'

        resp = self.carol.call_api(url_mapping, method=method, data=mapping_snapshot)
        _mapping_id = resp['mdmEntityMapping']['mdmId']

        if publish:
            url = f"v1/connectors/{connector_id}/entityMappings/{_mapping_id}/publish"
            self.carol.call_api(url, method='POST')


    def generate_SQL_tables(self, connector_id=None, connector_name=None):
        """ Generates SQL tables from a connector.

        The tables created will be named:
        `stg_{connector_id}_{staging_name}`. 

        Args:
            connector_id (str, optional): Connector ID. Defaults to None.
            connector_name (str, optional): Connector name. Defaults to None.

        Returns:
            dict: Carol's response
        

        """
        if connector_name:
            connector_id = self._connector_by_name(connector_name)
        elif connector_id is None:
            raise ValueError("Either `connector_id` or `connector_name` should be set") 

        params = {"connectorId": connector_id}
        return self.carol.call_api(path='v3/staging/generateSqlTables', method='POST', params=params, )


    def remove_SQL_tables(self, connector_id=None, connector_name=None):
        """Remove all SQL tables from a connector

        Args:
            connector_id (str, optional): Connector ID. Defaults to None.
            connector_name (str, optional): Connector name. Defaults to None.

        Returns:
            dict: Carol's response
        

        """

        raise NotImplementedError("API not inplemented in Carol yet. Talk for support.")


import os
import joblib
import luigi
import pandas as pd
from deprecated import deprecated

from pycarol.pipeline import LocalTarget


@deprecated
class PyCarolTarget(luigi.Target):
    """
    This is an abstract cloud target. Not to be called directly.
    In order to use PyCarol Targets, env or files configuration should allow Carol authentication
    with no parameters, Carol().
    If more than one tenant are used in same session, luigi parameter "tenant" should exist.
    """
    login_cache = None
    tenant_cache = None
    storage_cache = None

    is_cloud_target = True

    def __init__(self, task, *args, **kwargs):
        from ...carol import Carol
        from ...storage import Storage

        # We CANNOT cache the storage with GCP because the GCP API is not thread safe and would result in SSL errors
        # if luigi is using more than 1 worker
        if (PyCarolTarget.login_cache) and (PyCarolTarget.tenant_cache == task.tenant):
            self.login = PyCarolTarget.login_cache
            # self.storage = PyCarolTarget.storage_cache
        else:
            self.login = Carol()
            # self.storage = Storage(self.login)
            PyCarolTarget.login_cache = self.login
            # PyCarolTarget.storage_cache = self.storage
            PyCarolTarget.tenant_cache = task.tenant  # TODO: make cache more robust, not depending on task.tenant

        # Storage var needs to be always created per Target
        self.storage = Storage(self.login)

        namespace = task.get_task_namespace()
        file_id = task._file_id()
        file_id = file_id.split(namespace + '.')[
            -1]  # this will prevent to copy all the module path to the name of the file.
        self.path = os.path.join('pipeline', namespace, "{}.{}".format(file_id, self.FILE_EXT))
        self.log_path = os.path.join('pipeline', namespace, "{}_log.pkl".format(file_id))

    def persistlog(self, string):
        self.storage.save(self.log_path, string, format='joblib')

    def loadlog(self):
        try:
            text = self.storage.load(self.log_path, format='joblib')
        except Exception:
            return str(Exception)
        if not text:
            return "Log not found. log path: {}".format(self.log_path)
        return text

    def removelog(self):
        self.storage.delete(self.log_path)


@deprecated
class PyCarolFileTarget(PyCarolTarget):
    """
    This target operates with filepaths.
    easy_run should return a filepath for a local temporary file. This file will be removed after been sent to Carol.
    When loading this target, the file is copied from Carol to a local file. On easy_run we receive the local filepath.
    Important note: when loading the target, its local copy will not be automatically removed.
    """

    FILE_EXT = 'file'

    def load(self):
        return self.storage.load(self.path, format='file', cache=False)

    def dump(self, tempfile_path):
        self.storage.save(self.path, tempfile_path, format='file', cache=False)
        os.remove(tempfile_path)

    def remove(self):
        self.storage.delete(self.path)

    def exists(self):
        return self.storage.exists(self.path)


@deprecated
class PicklePyCarolTarget(PyCarolTarget):
    FILE_EXT = 'pkl'

    def load(self):
        return self.storage.load(self.path, format='joblib', cache=False)

    def dump(self, function_output):
        self.storage.save(self.path, function_output, format='joblib', cache=False)

    def remove(self):
        self.storage.delete(self.path)

    def exists(self):
        return self.storage.exists(self.path)


@deprecated
class ParquetPyCarolTarget(PyCarolTarget):
    FILE_EXT = 'parquet'

    def load(self, **kwargs):
        return self.storage.load(self.path, format='joblib', cache=True, parquet=True, **kwargs)

    def dump(self, function_output):
        self.storage.save(self.path, function_output, format='joblib', cache=False, parquet=True)

    def remove(self):
        self.storage.delete(self.path)

    def exists(self):
        return self.storage.exists(self.path)


@deprecated
class PytorchPyCarolTarget(PyCarolTarget):
    FILE_EXT = 'pth'

    def load(self):
        import torch
        local_path = self.storage.load(self.path, format='file')
        return torch.load(local_path)

    def dump(self, model_state_dict):
        import torch
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        torch.save(model_state_dict, self.path)
        self.storage.save(self.path, self.path, format='file')

    def remove(self):
        self.storage.delete(self.path)

    def exists(self):
        return self.storage.exists(self.path)


@deprecated
class KerasPyCarolTarget(PyCarolTarget):
    FILE_EXT = 'h5'

    def load(self):
        from keras.models import load_model
        local_path = self.storage.load(self.path, format='file')
        return load_model(local_path)

    def dump(self, model):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        model.save(self.path)
        self.storage.save(self.path, self.path, format='file')

    def remove(self):
        self.storage.delete(self.path)

    def exists(self):
        return self.storage.exists(self.path)


@deprecated
class PickleLocalTarget(LocalTarget):
    FILE_EXT = 'pkl'

    def load(self):
        return joblib.load(self.path)

    def dump(self, function_output):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        joblib.dump(function_output, self.path)

    def remove(self):
        try:
            os.remove(self.path)

        except(FileNotFoundError):
            print("file not found")


@deprecated
class ParquetLocalTarget(LocalTarget):
    FILE_EXT = 'parquet'

    def load(self, **kwargs):
        return pd.read_parquet(self.path, **kwargs)

    def dump(self, function_output):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        function_output.to_parquet(self.path, engine='fastparquet', has_nulls='infer')

    def remove(self):
        try:
            os.remove(self.path)
            print("file removed")
        except(FileNotFoundError):
            print("file not found")


@deprecated
class KerasLocalTarget(LocalTarget):
    FILE_EXT = 'h5'

    def load(self):
        from keras.models import load_model
        return load_model(self.path)

    def dump(self, model):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        model.save(self.path)

    def remove(self):
        try:
            os.remove(self.path)
            print("file removed")
        except(FileNotFoundError):
            print("file not found")


@deprecated
class PytorchLocalTarget(LocalTarget):
    FILE_EXT = 'pth'

    def load(self):
        import torch
        return torch.load(self.path)

    def dump(self, model_state_dict):
        import torch
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        torch.save(model_state_dict, self.path)

    def remove(self):
        try:
            os.remove(self.path)
            print("file removed")
        except(FileNotFoundError):
            print("file not found")


@deprecated
class JsonLocalTarget(LocalTarget):
    FILE_EXT = 'json'

    def load(self):
        return pd.read_json(self.path)

    def dump(self, function_output):
        # TODO: json only works for dataframe
        function_output.to_json(self.path)

    def remove(self):
        os.remove(self.path)


@deprecated
class FeatherLocalTarget(LocalTarget):
    FILE_EXT = 'feather'

    def load(self):
        import feather
        return feather.read_dataframe(self.path)

    def dump(self, function_output):
        import feather
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        feather.write_dataframe(function_output, self.path)

    def remove(self):
        os.remove(self.path)



from pycarol import Carol, Staging, Connectors, DataModel
from collections import defaultdict
import itertools


def find_keys(node, kv):
    """Find recursively all the values from a given key. 

    Args:
        node (dict of dict): Nested dictionary.
        kv (str): dictionary key to find. 

    Yields:
        dict: dictionary
    """
    if isinstance(node, list):
        for i in node:
            yield from find_keys(i, kv)
    elif isinstance(node, dict):
        if kv in node:
            yield node[kv]
        for j in node.values():
            yield from find_keys(j, kv)


def unroll_list(l):
    """Unroll a list of lists to a flat list

    Args:
        l (list of lists): List of list to unroll

    Yields:
        list: unrolled list.
    """

    if isinstance(l, list):
        for i in l:
            yield from unroll_list(i)
    else:
        yield l


def get_staging_prefix(carol, staging_prefix, connector_name, connector_id):
    """Create the prefix for a given set of staging_prefix, connector_name, connector_id

    Args:
        carol: `pycarol.Carol`
            Carol() object.
        connector_name: `str`
            Connector Name
        prefix: 'str` default `DM_`
            prefix to add to the data model name. e.g., if dm_name='mydatamoldel', the result will be "DM_mydatamoldel`.
            This is only applied for DataModel.
        staging_prefix: 'str` default `None`
            Prefix for the staging name. e.g., if staging_prefix='connector_name` the output will be:
                    { "connector_name_stag3" : {"connector_name_stag1", "connector_name_stag2"}}
            Possible values are: 'connector_name', 'connector_id', None

    """
    conn = Connectors(carol)
    if staging_prefix is None:
        staging_prefix = ''
    elif staging_prefix == 'connector_name':
        if connector_name is None:
            connector_name = conn.get_by_id(connector_id)['mdmName']
        staging_prefix = connector_name + '_'
    elif staging_prefix == 'connector_id':
        if connector_id is None:
            connector_id = conn.get_by_name(connector_name)['mdmId']
        staging_prefix = connector_id + '_'
    else:
        raise ValueError(
            f'`staging_prefix` should be either `connector_name`, `connector_id` or `None` {staging_prefix} was used'
        )
    return staging_prefix, connector_name, connector_id


def get_dm_relationship_constraints(carol, connector_name=None, connector_id=None, dm_prefix='DM_', staging_prefix=None):
    """
    Create relationship between data models based on their relationship constraints

    Args:
        carol: `pycarol.Carol`
            Carol() object.
        connector_name: `str`
            Connector Name
        dm_prefix: 'str` default `DM_`
            data model prefix to add to the data model name. e.g., if dm_name='mydatamoldel', the result will be "DM_mydatamoldel`

    Returns: `defaultdict(set)`
        dictionary { "dm1" : {"dm2", "dm3"}} where "dm1" depends on "dm2"" and "dm3"

    """
    staging_prefix, connector_name, connector_id = get_staging_prefix(
        carol, staging_prefix=staging_prefix,
        connector_name=connector_name, connector_id=connector_id
    )

    conn = Connectors(carol)
    conn_to_create = conn.get_all(
        include_mappings=True, include_connectors=True, include_consumption=True)
    conn_to_create = [
        i for i in conn_to_create if i['mdmName'] == connector_name]
    dm_mapping = defaultdict(set)
    for staging_name, mapping in conn_to_create[0]['mdmEntityMappings'].items():
        dm_mapping[dm_prefix + mapping['mdmMasterEntityName']].update([staging_name])

    # find Relationship Constraints
    dm = DataModel(carol)
    dms = dm.get_all().template_dict.keys()

    relationship_constraints = defaultdict(set)
    for i in dms:
        snap = dm.get_by_name(i)['mdmRelationshipConstraints']
        relationship_constraints[dm_prefix + i].update([dm_prefix + j['mdmTargetEntityName'] for j in snap])
        # if there are more than one staging mapping to a DM I need to restrict them too, sinse process staging mean
        # create golden records.
        for parent_staging in dm_mapping[dm_prefix + i]:
            relationship_constraints[staging_prefix+parent_staging].update(
                [dm_prefix + j['mdmTargetEntityName'] for j in snap])

    return relationship_constraints


def get_mapping_constraints(carol, connector_name=None, connector_id=None, dm_prefix='DM_', staging_prefix=None):
    """
    Create relationship between data models and stagings in mappings.

    Args:
        carol: `pycarol.Carol`
            Carol() object.
        connector_name: `str`
            Connector Name
        connector_id: `str`
            Connector ID
        dm_prefix: 'str` default `DM_`
            data model prefix to add to the data model name. e.g., if dm_name='mydatamoldel', the result will be "DM_mydatamoldel`.
            This is only applied for DataModel.
        staging_prefix: 'str` default `None`
            Prefix for the staging name. e.g., if staging_prefix='connector_name` the output will be:
                    { "connector_name_stag3" : {"connector_name_stag1", "connector_name_stag2"}}
            Possible values are: 'connector_name', 'connector_id', None

    Returns: `defaultdict(set)`
        dictionary { "dm1" : {"stag1", "stag2"}} where "dm1" depends on "stag1"" and "stag2"

    """

    lookup_types = ['LOOKUP_DATA_MODEL', 'LOOKUP_STAGING_TABLE', 'LOOKUP_MULTIPLE_DATA_MODELS',
                    'LOOKUP_MULTIPLE_STAGING_TABLES']

    staging_prefix, connector_name, connector_id = get_staging_prefix(
        carol, staging_prefix=staging_prefix,
        connector_name=connector_name, connector_id=connector_id
    )

    conn = Connectors(carol)
    conn_to_create = conn.get_all(
        include_mappings=True, include_connectors=True, include_consumption=True)
    conn_to_create = [
        i for i in conn_to_create if i['mdmName'] == connector_name]

    stag = Staging(carol)

    dm_to_mapping = defaultdict(set)
    for staging_source in conn_to_create[0]['mdmEntityMappings'].keys():
        # staging_source = 'currency'
        curr_map = conn_to_create[0]['mdmEntityMappings'][staging_source]
        assert staging_source == curr_map['mdmStagingType']
        staging_source = curr_map['mdmStagingType']
        dm_target = curr_map['mdmMasterEntityName']
        mapping_id = curr_map['mdmId']
        entity_space = curr_map['mdmEntitySpace']
        connector_id = curr_map['mdmConnectorId']
        mappings_to_get = stag.get_mapping_snapshot(connector_id=connector_id, mapping_id=mapping_id,
                                                    entity_space=entity_space)
        dm_to_mapping[dm_prefix +
                      dm_target].update([staging_prefix + staging_source])

        # Get mapping lookups
        for i in mappings_to_get[None]['fieldMappings']:
            field_cleanse_rules = i['fieldCleanseRules']
            for cleanse in field_cleanse_rules:

                # it can be inside actions and falseActions if inside an If.
                unest_func = list(itertools.chain(*(list(find_keys(cleanse, 'actions'))
                                                    + list(find_keys(cleanse, 'falseActions'))
                                                    + [[cleanse, ]])))

                for cleanse_func in unest_func:
                    if cleanse_func['fieldFunction']['mdmName'] in lookup_types:
                        if "DATA_MODEL" in cleanse_func['fieldFunction']['mdmName']:
                            _prefix = dm_prefix
                            # index 0 sinde it is the only parameter for DM lookups.
                            idx = 0
                        elif "STAGING_TABLE" in cleanse_func['fieldFunction']['mdmName']:
                            # index 1 sinde the first param is the connector name.
                            idx = 1
                            _prefix = staging_prefix

                        dm_to_mapping[staging_prefix + staging_source].update(
                            [_prefix + cleanse_func['parameterValues'][idx]])

    return dm_to_mapping


def get_etl_constraints(carol, connector_name=None, connector_id=None, staging_prefix=None):
    """
    Create relationship between stagings in ETLs.

    Args:
        carol: `pycarol.Carol`
            Carol() object.
        connector_name: `str`
            connector name to get the relationships.
        connector_id: `str`
            connector ID to get the relationships.
        staging_prefix: 'str` default `None`
            Prefix for the staging name. e.g., if staging_prefix='connector_name` the output will be:
                 { "connector_name_stag3" : {"connector_name_stag1", "connector_name_stag2"}}
            Possible values are: 'connector_name', 'connector_id', None

    Returns: `defaultdict(set)`
        dictionary { "stag3" : {"stag1", "stag2"}} where "stag3" depends on "stag1"" and "stag2"

    """

    if connector_id is None and connector_name is None:
        raise ValueError(
            'Either connector_id or connector_name must be set.')

    conn = Connectors(carol)

    staging_prefix, connector_name, connector_id = get_staging_prefix(
        carol, staging_prefix=staging_prefix,
        connector_name=connector_name, connector_id=connector_id
    )

    st = conn.get_all_stagings(
        connector_name=connector_name, connector_id=connector_id)

    etls = conn.get_etl_information(
        connector_name=connector_name, connector_id=connector_id)
    rel_etls = defaultdict(set)
    for etl in etls:
        if etl.get("mdmETLType", "") == 'JOIN':
            source = staging_prefix + etl['mdmSourceEntityName']
            other_side = staging_prefix + etl['mdmOtherSourceEntityName']
            etl_target = list(find_keys(etl, 'mdmParameterValues'))
            etl_target = list(unroll_list(etl_target))

            # get the target table.
            assert len(etl_target) == 1
            etl_target = etl_target[0]

            # get source.
            etl_source = set(itertools.chain(
                *[i.keys() for i in find_keys(etl, 'mdmStagingTypeToFieldForMatching')]))
            etl_source = [staging_prefix + i for i in etl_source]
            rel_etls[staging_prefix + etl_target].update(etl_source)
            # Need to add the sourve->other side to the list. We need to guarantee that the other side is there to do the join.
            rel_etls[source].update({other_side})

        elif etl.get("mdmETLType", "") in ['SPLIT', 'DUPLICATE']:
            # get all actions
            source = staging_prefix + etl['mdmSourceEntityName']
            split_actions = list(find_keys(etl, 'mdmParameterValues'))
            split_actions = set(unroll_list(split_actions))
            for target in split_actions:
                if target in st:
                    rel_etls[staging_prefix+target].update({source})

        else:
            raise ValueError(
                f'mdmETLType are "SPLIT", "DUPLICATE" or "JOIN", {etl.get("mdmETLType", "")} were received')

    return rel_etls


def generate_dependency_graph(carol, connector_list=None, dm_prefix='DM_', staging_prefix='connector_name', only_mapping=False):
    """Generates dependency graph for a list of connectors.

    Args:
        carol (pycarol.Carol): INstance of Carol
        connector_list (list, optional): List of connectors to create the dependency graph. Defaults to None.
        dm_prefix: 'str` default `DM_`
            data model prefix to add to the data model name. e.g., if dm_name='mydatamoldel', the result will be "DM_mydatamoldel`.
            This is only applied for DataModel.
        staging_prefix: 'str` default `None`
            Prefix for the staging name. e.g., if staging_prefix='connector_name` the output will be:
                 { "connector_name_stag3" : {"connector_name_stag1", "connector_name_stag2"}}
            Possible values are: 'connector_name', 'connector_id', None
        only_mapping (bool, optional): Only generates mappings and constraints dependency graph. Defaults to False.

    Returns:
        default: defaultdict(set). Dependency graph.
    """

    if connector_list is None:
        connector_list = Connectors(carol).get_all()
        connector_list = [i['mdmName'] for i in connector_list]

    
    all_rel = defaultdict(set)
    for connector_name in connector_list:

        dm_constraints = get_dm_relationship_constraints(carol, dm_prefix=dm_prefix, connector_name=connector_name, )
        mapping_constraint = get_mapping_constraints(carol, connector_name=connector_name, dm_prefix=dm_prefix, staging_prefix=staging_prefix)
        if not only_mapping:
            etl_constraints = get_etl_constraints(carol, connector_name=connector_name,  staging_prefix=staging_prefix )
        else:
            etl_constraints = defaultdict(set)


        for d in [dm_constraints, etl_constraints, mapping_constraint]:
            for key, value in d.items():
                all_rel[key].update(all_rel[key] | value)

    return all_rel


import multiprocessing
import functools
import warnings
import gzip
import json

from tqdm import tqdm

__STAGING_FIELDS = ['mdmCounterForEntity', 'mdmId']
__DM_FIELDS = ['mdmCounterForEntity', 'mdmId']


def _import_dask(
        storage, merge_records=False,
        dm_name=None, import_type='staging', return_dask_graph=False,
        connector_id=None, staging_name=None, view_name=None, columns=None,
        max_hits=None, mapping_columns=None, engine='pyarrow', file_pattern=None,
):
    file_pattern = file_pattern if file_pattern else ''
    from dask import dataframe as dd
    mapping_columns = mapping_columns or {}  # dask does not accepet None. Need to send a valid mapping.

    if columns:
        columns = list(set(columns))
        columns += __STAGING_FIELDS
        columns = list(set(columns))

    is_parquet = True
    if import_type == 'golden':
        url = [storage.build_url_parquet_golden(dm_name=dm_name)]
    elif import_type == 'golden_cds':
        url = storage.build_url_parquet_golden_cds(dm_name=dm_name)
    elif import_type == 'staging':
        url = []
        url1 = storage.build_url_parquet_staging(staging_name=staging_name, connector_id=connector_id)
        if url1 is not None:
            url.append(url1)
        url2 = storage.build_url_parquet_staging_master(staging_name=staging_name, connector_id=connector_id)
        if url2 is not None:
            url.append(url2)
        url3 = storage.build_url_parquet_staging_rejected(staging_name=staging_name, connector_id=connector_id)
        if url3 is not None:
            url.append(url3)
    elif import_type == 'view':
        url = [storage.build_url_parquet_view(view_name=view_name)]
    elif import_type == 'staging_cds':
        url = storage.build_url_parquet_staging_cds(staging_name=staging_name, connector_id=connector_id)
    elif import_type == 'golden_rejected':
        url = storage.build_url_parquet_golden_rejected_cds(dm_name=dm_name)
    else:
        raise KeyError('import_type should be `golden`,`staging`, `view`, `staging_cds`, `golden_cds`, `view_cds`,'
                       '`golden_rejected`')

    url = url + file_pattern + "*.parquet"
    d = dd.read_parquet(url, storage_options=storage.get_dask_options(), columns=columns, engine=engine)
    d = d.rename(columns=mapping_columns)
    if return_dask_graph:
        return d
    else:
        return d.compute()


def _import_pandas(storage, dm_name=None, connector_id=None, columns=None, mapping_columns=None, max_workers=None,
                   staging_name=None, view_name=None, import_type='staging', golden=False, max_hits=None, callback=None,
                   token_carolina=None, storage_space=None, file_pattern=None):
    import pandas as pd

    if callback is not None and not callable(callback):
        raise TypeError(f'"{callback}" is a {type(callback)} and is not callable.')

    if columns:
        columns = list(set(columns))
        columns += __DM_FIELDS
        columns = list(set(columns))

    if import_type == 'golden':
        file_paths = storage.get_golden_file_paths(dm_name=dm_name)
    elif import_type == 'staging':
        file_paths = storage.get_staging_file_paths(staging_name=staging_name, connector_id=connector_id)
    elif import_type == 'view':
        file_paths = storage.get_view_file_paths(view_name=view_name)
    elif import_type == 'staging_cds':
        file_paths = storage.get_staging_cds_file_paths(staging_name=staging_name, connector_id=connector_id,
                                                        file_pattern=file_pattern)
    elif import_type == 'golden_cds':
        file_paths = storage.get_golden_cds_file_paths(dm_name=dm_name,
                                                       file_pattern=file_pattern)
    elif import_type == 'view_cds':
        file_paths = storage.get_view_cds_file_paths(dm_name=view_name)
    elif import_type == 'golden_rejected':
        file_paths = storage.get_golden_rejected_cds_file_paths(dm_name=dm_name, file_pattern=file_pattern)
    else:
        raise KeyError('import_type should be `golden`,`staging`, `view`, `staging_cds`, `golden_cds`, `view_cds`',
                       'golden_rejected')

    df_list = []
    count = 0

    if max_workers is not None:
        assert max_workers > 0, f"max_workers must be greater than zero, you passed {max_workers}"
    else:
        max_workers = 1

    if max_workers > 1:
        client = _load_client(token_carolina)
        if max_hits:
            warnings.warn("max_hits does not work when max_hits>1", DeprecationWarning)
        partial_download = functools.partial(_download_files, storage=client, storage_space=storage_space,
                                             columns=columns,
                                             mapping_columns=mapping_columns, callback=callback)
        with multiprocessing.Pool(processes=max_workers) as pool:
            df_list = pool.map(partial_download, file_paths)
    else:
        for i, file in enumerate(tqdm(list(file_paths))):  # need list to be able to track the counts.
            buffer = storage.load(file['name'], format='raw', cache=False, storage_space=file['storage_space'])
            if file['name'].endswith('.parquet'):
                result = pd.read_parquet(buffer, columns=columns)
            else:
                ValueError('Supported files are `parquet`')

            if mapping_columns is not None:
                # fix columns names (we replace `-` for `_` due to parquet limitations.
                result.rename(columns=mapping_columns, inplace=True)
            if callback:
                result = callback(result)

            df_list.append(result)
            if max_hits is not None:
                count_old = count
                count += len(df_list[i])
                if count >= max_hits:
                    df_list[i] = df_list[i].iloc[:max_hits - count_old]
                    break
    if not df_list:
        return None
    return pd.concat(df_list, ignore_index=True, sort=True)


def _download_files(file, storage, storage_space, columns, mapping_columns, callback):
    import pandas as pd
    filename = storage_space + '/' + file['name']
    buffer = storage.open(filename)
    if file['name'].endswith('.parquet'):
        result = pd.read_parquet(buffer, columns=columns)
    else:
        ValueError('Supported files are `parquet`')
    if mapping_columns is not None:
        result.rename(columns=mapping_columns, inplace=True)
    if callback:
        result = callback(result)
    return result


def _load_client(token):
    import gcsfs
    client = gcsfs.GCSFileSystem(token=token)
    return client
